from collections import defaultdict
from pathlib import Path
from typing import Final
import logging

import numpy

from ptychodus.api.plugins import PluginRegistry
from ptychodus.api.probe_positions import (
    ProbePositionSequence,
    ProbePositionFileReader,
    ProbePosition,
)

logger = logging.getLogger(__name__)


class APS12IDPositionFileReader(ProbePositionFileReader):
    ONE_NANOMETER_M: Final[float] = 1.0e-9

    def read(self, file_path: Path) -> ProbePositionSequence:
        stem_parts = file_path.stem.split('_')
        stem_prefix = '_'.join(stem_parts[:-2])
        logger.debug(f'{stem_prefix=}')
        scan_num = int(stem_parts[-3])
        logger.debug(f'{scan_num=}')

        lines: set[int] = set()
        points: set[int] = set()
        points_per_line: dict[int, set[int]] = defaultdict(set[int])
        file_dict: dict[tuple[int, int], Path] = dict()

        for p in file_path.parent.glob(f'{stem_prefix}_*{file_path.suffix}'):
            stem_parts = p.stem.split('_')
            line = int(stem_parts[-2])
            point = int(stem_parts[-1])

            lines.add(line)
            points.add(point)
            points_per_line[line].add(point)
            file_dict[line, point] = p

        logger.debug(f'{lines=}')
        lines_min = min(lines)
        # lines_max = max(lines)
        # lines_num = lines_max - lines_min + 1
        logger.debug(f'{points=}')
        points_min = min(points)
        points_max = max(points)
        points_num = points_max - points_min + 1

        for line, line_points in points_per_line.items():
            missing_points = points - line_points

            if missing_points:
                logger.warning(f'Line {line} is missing points {missing_points}')

        scan_point_list: list[ProbePosition] = list()

        for (line, point), fp in sorted(file_dict.items()):
            index = (point - points_min) + (line - lines_min) * points_num
            position_data = numpy.genfromtxt(fp)

            for row in position_data:
                scan_point = ProbePosition(
                    index=index,
                    coordinate_x_m=-self.ONE_NANOMETER_M * row[2],
                    coordinate_y_m=+self.ONE_NANOMETER_M * row[1],
                )
                scan_point_list.append(scan_point)

        return ProbePositionSequence(scan_point_list)


def register_plugins(registry: PluginRegistry) -> None:
    registry.probe_position_file_readers.register_plugin(
        APS12IDPositionFileReader(),
        simple_name='APS_PtychoSAXS',
        display_name='APS 12-ID PtychoSAXS Files (*.dat)',
    )



from decimal import Decimal
from typing import Any, Final
import logging

from PyQt5.QtCore import Qt, QAbstractListModel, QModelIndex, QObject, QStringListModel
from PyQt5.QtWidgets import QWidget

from ptychodus.api.observer import Observable, Observer

from ...model.fluorescence import (
    FluorescenceEnhancer,
    TwoStepFluorescenceEnhancingAlgorithm,
    VSPIFluorescenceEnhancingAlgorithm,
)
from ...model.visualization import VisualizationEngine
from ...view.probe import (
    FluorescenceDialog,
    FluorescenceTwoStepParametersView,
    FluorescenceVSPIParametersView,
)
from ...view.widgets import ExceptionDialog
from ..data import FileDialogFactory
from ..helpers import connect_current_changed_signal
from ..visualization import (
    VisualizationParametersController,
    VisualizationWidgetController,
)

logger = logging.getLogger(__name__)


class FluorescenceChannelListModel(QAbstractListModel):
    def __init__(self, enhancer: FluorescenceEnhancer, parent: QObject | None = None) -> None:
        super().__init__(parent)
        self._enhancer = enhancer

    def data(self, index: QModelIndex, role: int = Qt.ItemDataRole.DisplayRole) -> Any:
        # TODO make this a table model and show measured/enhanced count statistics
        if index.isValid() and role == Qt.ItemDataRole.DisplayRole:
            emap = self._enhancer.get_measured_element_map(index.row())
            return emap.name

    def rowCount(self, parent: QModelIndex = QModelIndex()) -> int:  # noqa: N802
        return self._enhancer.get_num_channels()


class FluorescenceTwoStepViewController(Observer):
    def __init__(self, algorithm: TwoStepFluorescenceEnhancingAlgorithm) -> None:
        super().__init__()
        self._algorithm = algorithm
        self._view = FluorescenceTwoStepParametersView()

        self._upscaling_model = QStringListModel()
        self._upscaling_model.setStringList(self._algorithm.get_upscaling_strategies())
        self._view.upscaling_strategy_combo_box.setModel(self._upscaling_model)
        self._view.upscaling_strategy_combo_box.textActivated.connect(
            algorithm.set_upscaling_strategy
        )

        self._deconvolution_model = QStringListModel()
        self._deconvolution_model.setStringList(self._algorithm.get_deconvolution_strategies())
        self._view.deconvolution_strategy_combo_box.setModel(self._deconvolution_model)
        self._view.deconvolution_strategy_combo_box.textActivated.connect(
            algorithm.set_deconvolution_strategy
        )

        self._sync_model_to_view()
        algorithm.add_observer(self)

    def get_widget(self) -> QWidget:
        return self._view

    def _sync_model_to_view(self) -> None:
        self._view.upscaling_strategy_combo_box.setCurrentText(
            self._algorithm.get_upscaling_strategy()
        )
        self._view.deconvolution_strategy_combo_box.setCurrentText(
            self._algorithm.get_deconvolution_strategy()
        )

    def _update(self, observable: Observable) -> None:
        if observable is self._algorithm:
            self._sync_model_to_view()


class FluorescenceVSPIViewController(Observer):
    MAX_INT: Final[int] = 0x7FFFFFFF

    def __init__(self, algorithm: VSPIFluorescenceEnhancingAlgorithm) -> None:
        super().__init__()
        self._algorithm = algorithm
        self._view = FluorescenceVSPIParametersView()

        self._view.damping_factor_line_edit.value_changed.connect(
            self._sync_damping_factor_to_model
        )
        self._view.max_iterations_spin_box.setRange(1, self.MAX_INT)
        self._view.max_iterations_spin_box.valueChanged.connect(algorithm.set_max_iterations)

        algorithm.add_observer(self)
        self._sync_model_to_view()

    def get_widget(self) -> QWidget:
        return self._view

    def _sync_damping_factor_to_model(self, value: Decimal) -> None:
        self._algorithm.set_damping_factor(float(value))

    def _sync_model_to_view(self) -> None:
        self._view.damping_factor_line_edit.set_value(
            Decimal(repr(self._algorithm.get_damping_factor()))
        )
        self._view.max_iterations_spin_box.setValue(self._algorithm.get_max_iterations())

    def _update(self, observable: Observable) -> None:
        if observable is self._algorithm:
            self._sync_model_to_view()


class FluorescenceViewController(Observer):
    def __init__(
        self,
        enhancer: FluorescenceEnhancer,
        engine: VisualizationEngine,
        file_dialog_factory: FileDialogFactory,
    ) -> None:
        super().__init__()
        self._enhancer = enhancer
        self._engine = engine
        self._file_dialog_factory = file_dialog_factory
        self._dialog = FluorescenceDialog()
        self._enhancement_model = QStringListModel()
        self._enhancement_model.setStringList(self._enhancer.algorithms())
        self._channel_list_model = FluorescenceChannelListModel(enhancer)

        self._dialog.fluorescence_parameters_view.open_button.clicked.connect(
            self._open_measured_dataset
        )

        two_step_view_controller = FluorescenceTwoStepViewController(
            enhancer.two_step_enhancing_algorithm
        )
        self._dialog.fluorescence_parameters_view.algorithm_combo_box.addItem(
            TwoStepFluorescenceEnhancingAlgorithm.DISPLAY_NAME,
            self._dialog.fluorescence_parameters_view.algorithm_combo_box.count(),
        )
        self._dialog.fluorescence_parameters_view.stacked_widget.addWidget(
            two_step_view_controller.get_widget()
        )

        vspi_view_controller = FluorescenceVSPIViewController(enhancer.vspi_enhancing_algorithm)
        self._dialog.fluorescence_parameters_view.algorithm_combo_box.addItem(
            VSPIFluorescenceEnhancingAlgorithm.DISPLAY_NAME,
            self._dialog.fluorescence_parameters_view.algorithm_combo_box.count(),
        )
        self._dialog.fluorescence_parameters_view.stacked_widget.addWidget(
            vspi_view_controller.get_widget()
        )

        self._dialog.fluorescence_parameters_view.algorithm_combo_box.textActivated.connect(
            enhancer.set_algorithm
        )
        self._dialog.fluorescence_parameters_view.algorithm_combo_box.currentIndexChanged.connect(
            self._dialog.fluorescence_parameters_view.stacked_widget.setCurrentIndex
        )
        self._dialog.fluorescence_parameters_view.algorithm_combo_box.setModel(
            self._enhancement_model
        )
        self._dialog.fluorescence_parameters_view.algorithm_combo_box.textActivated.connect(
            enhancer.set_algorithm
        )

        self._dialog.fluorescence_parameters_view.enhance_button.clicked.connect(
            self._enhance_fluorescence
        )
        self._dialog.fluorescence_parameters_view.save_button.clicked.connect(
            self._save_enhanced_dataset
        )

        self._dialog.fluorescence_channel_list_view.setModel(self._channel_list_model)
        connect_current_changed_signal(
            self._dialog.fluorescence_channel_list_view, self._update_view
        )

        self._measured_widget_controller = VisualizationWidgetController(
            engine,
            self._dialog.measured_widget,
            self._dialog.status_bar,
            file_dialog_factory,
        )
        self._enhanced_widget_controller = VisualizationWidgetController(
            engine,
            self._dialog.enhanced_widget,
            self._dialog.status_bar,
            file_dialog_factory,
        )
        self._visualization_parameters_controller = VisualizationParametersController(
            engine, self._dialog.visualization_parameters_view
        )

        enhancer.add_observer(self)

    def _open_measured_dataset(self) -> None:
        title = 'Open Measured Fluorescence Dataset'
        file_path, name_filter = self._file_dialog_factory.get_open_file_path(
            self._dialog,
            title,
            name_filters=[nf for nf in self._enhancer.get_open_file_filters()],
            selected_name_filter=self._enhancer.get_open_file_filter(),
        )

        if file_path:
            try:
                self._enhancer.open_measured_dataset(file_path, name_filter)
            except Exception as err:
                logger.exception(err)
                ExceptionDialog.show_exception(title, err)

    def _enhance_fluorescence(self) -> None:
        try:
            self._enhancer.enhance_fluorescence()
        except Exception as err:
            logger.exception(err)
            ExceptionDialog.show_exception('Enhance Fluorescence', err)

    def launch(self, product_index: int) -> None:
        self._enhancer.set_product(product_index)

        try:
            item_name = self._enhancer.get_product_name()
        except Exception as err:
            logger.exception(err)
            ExceptionDialog.show_exception('Launch', err)
        else:
            self._dialog.setWindowTitle(f'Enhance Fluorescence: {item_name}')
            self._dialog.open()

    def _save_enhanced_dataset(self) -> None:
        title = 'Save Enhanced Fluorescence Dataset'
        file_path, name_filter = self._file_dialog_factory.get_save_file_path(
            self._dialog,
            title,
            name_filters=[nf for nf in self._enhancer.get_save_file_filters()],
            selected_name_filter=self._enhancer.get_save_file_filter(),
        )

        if file_path:
            try:
                self._enhancer.save_enhanced_dataset(file_path, name_filter)
            except Exception as err:
                logger.exception(err)
                ExceptionDialog.show_exception(title, err)

    def _sync_model_to_view(self) -> None:
        self._dialog.fluorescence_parameters_view.algorithm_combo_box.setCurrentText(
            self._enhancer.get_algorithm()
        )
        self._channel_list_model.beginResetModel()
        self._channel_list_model.endResetModel()

    def _update_view(self, current: QModelIndex, previous: QModelIndex) -> None:
        if not current.isValid():
            self._measured_widget_controller.clear_array()
            self._enhanced_widget_controller.clear_array()
            return

        try:
            emap_measured = self._enhancer.get_measured_element_map(current.row())
        except Exception as err:
            logger.exception(err)
            self._measured_widget_controller.clear_array()
            ExceptionDialog.show_exception('Render Measured Element Map', err)
        else:
            self._measured_widget_controller.set_array(
                emap_measured.counts_per_second, self._enhancer.get_pixel_geometry()
            )

        try:
            emap_enhanced = self._enhancer.get_enhanced_element_map(current.row())
        except Exception as err:
            logger.exception(err)
            self._enhanced_widget_controller.clear_array()
            ExceptionDialog.show_exception('Render Enhanced Element Map', err)
        else:
            self._enhanced_widget_controller.set_array(
                emap_enhanced.counts_per_second, self._enhancer.get_pixel_geometry()
            )

    def _update(self, observable: Observable) -> None:
        if observable is self._enhancer:
            self._sync_model_to_view()



from collections.abc import Iterator, Sequence


class PtychoPINNEnumerators:
    def __init__(self) -> None:
        self._amp_activations: Sequence[str] = ['sigmoid', 'swish', 'softplus', 'relu']

    def get_amp_activations(self) -> Iterator[str]:
        return iter(self._amp_activations)



from ptychodus.api.observer import Observable, Observer
from ptychodus.api.settings import SettingsRegistry


class ProcessingSettings(Observable, Observer):
    def __init__(self, registry: SettingsRegistry) -> None:
        super().__init__()
        self._group = registry.create_group('Processing')
        self._group.add_observer(self)

        self.algorithm = self._group.create_string_parameter('Algorithm', 'pty-chi_LSQML')

    def _update(self, observable: Observable) -> None:
        if observable is self._group:
            self.notify_observers()



from pathlib import Path
from typing import Final
import logging
import sys

from scipy.stats import gaussian_kde
import h5py
import matplotlib.pyplot as plt
import numpy

from ptychodus.api.plugins import PluginRegistry
from ptychodus.api.probe_positions import (
    ProbePositionSequence,
    ProbePositionFileReader,
    ProbePosition,
)

# use full module path to make this file usable as an entry point
from ptychodus.plugins.h5_diffraction_file import H5DiffractionFileReader

logger = logging.getLogger(__name__)


class LCLSPositionFileReader(ProbePositionFileReader):
    SIMPLE_NAME: Final[str] = 'LCLS_XPP'  # noqa: N806
    DISPLAY_NAME: Final[str] = 'LCLS X-ray Pump Probe Files (*.h5 *.hdf5)'  # noqa: N806
    ONE_MICRON_M: Final[float] = 1e-6

    def __init__(
        self,
        tomography_angle_deg: float = 180.0,
        ipm2_low_threshold: float = 2000.0,
        ipm2_high_threshold: float = 3000.0,
    ) -> None:
        self._tomography_angle_deg = tomography_angle_deg
        self._ipm2_low_threshold = ipm2_low_threshold
        self._ipm2_high_threshold = ipm2_high_threshold

    def read(self, file_path: Path) -> ProbePositionSequence:
        point_list: list[ProbePosition] = list()

        with h5py.File(file_path, 'r') as h5_file:
            # piezo stage positions are in microns
            piezo_stage_position_x_um = h5_file['/lmc/ch03'][:]
            piezo_stage_position_y_um = h5_file['/lmc/ch04'][:]
            piezo_stage_position_z_um = h5_file['/lmc/ch05'][:]

            # ipm2 is used for filtering and normalizing the data
            ipm2 = h5_file['/ipm2/sum'][:]

            # vertical coordinate
            ycoords = -piezo_stage_position_z_um * self.ONE_MICRON_M

            # horizontal coordinate
            tomography_angle_rad = numpy.deg2rad(self._tomography_angle_deg)
            cos_angle = numpy.cos(tomography_angle_rad)
            sin_angle = numpy.sin(tomography_angle_rad)
            xcoords = (
                cos_angle * piezo_stage_position_x_um + sin_angle * piezo_stage_position_y_um
            ) * self.ONE_MICRON_M

            for index, (ipm, x, y) in enumerate(zip(ipm2, xcoords, ycoords)):
                if self._ipm2_low_threshold <= ipm and ipm < self._ipm2_high_threshold:
                    if numpy.isfinite(x) and numpy.isfinite(y):
                        point = ProbePosition(index, x, y)
                        point_list.append(point)
                else:
                    logger.debug(f'Filtered scan point {index=} {ipm=}.')

        return ProbePositionSequence(point_list)


def register_plugins(registry: PluginRegistry) -> None:
    registry.diffraction_file_readers.register_plugin(
        H5DiffractionFileReader(data_path='/jungfrau1M/image_img'),
        simple_name=LCLSPositionFileReader.SIMPLE_NAME,
        display_name=LCLSPositionFileReader.DISPLAY_NAME,
    )
    registry.probe_position_file_readers.register_plugin(
        LCLSPositionFileReader(),
        simple_name=LCLSPositionFileReader.SIMPLE_NAME,
        display_name=LCLSPositionFileReader.DISPLAY_NAME,
    )


if __name__ == '__main__':
    file_path = Path(sys.argv[1])

    with h5py.File(file_path, 'r') as h5_file:
        ipm2 = h5_file['/ipm2/sum'][:]

    q0 = numpy.min(ipm2)
    q1, q2, q3 = numpy.quantile(ipm2, [0.25, 0.50, 0.75])
    q4 = numpy.max(ipm2)

    print(f'ipm2: {ipm2.dtype}{ipm2.shape}')
    print(f'min = {q0}')
    print(f'Q1 = {q1}')
    print(f'Q2 = {q2}')
    print(f'Q3 = {q3}')
    print(f'max = {q4}')

    fig = plt.figure()
    ax = fig.add_subplot(111)
    ax.plot(ipm2, '.', linewidth=1.5)
    ax.set_title('IPM2')
    ax.set_xlabel('Position')
    ax.set_ylabel('Value')
    ax.grid(True)
    plt.show()

    values = numpy.linspace(q0, q4, 1000)
    kde = gaussian_kde(ipm2)
    density = kde(values)

    fig = plt.figure()
    ax = fig.add_subplot(111)
    ax.plot(values, density, '.-', linewidth=1.5)
    ax.set_title('IPM2')
    ax.set_xlabel('Value')
    ax.set_ylabel('Density')
    ax.grid(True)
    plt.show()



from .core import AutomationCore, AutomationPresenter
from .repository import AutomationDataset, AutomationDatasetRepository
from .repository import AutomationDatasetState
from .settings import AutomationSettings

__all__ = [
    'AutomationCore',
    'AutomationDataset',
    'AutomationDatasetRepository',
    'AutomationDatasetState',
    'AutomationPresenter',
    'AutomationSettings',
]



from __future__ import annotations
from abc import abstractmethod
from collections.abc import Sequence
from pathlib import Path
import logging

from ptychodus.api.parametric import ParameterGroup
from ptychodus.api.probe_positions import (
    ProbePositionSequence,
    ProbePositionFileReader,
    ProbePosition,
)

from .settings import ProbePositionsSettings

logger = logging.getLogger(__name__)


class ProbePositionsBuilder(ParameterGroup):
    def __init__(self, settings: ProbePositionsSettings, name: str) -> None:
        super().__init__()
        self._name = settings.builder.copy()
        self._name.set_value(name)
        self._add_parameter('name', self._name)

    def get_name(self) -> str:
        return self._name.get_value()

    def sync_to_settings(self) -> None:
        for parameter in self.parameters().values():
            parameter.sync_value_to_parent()

    @abstractmethod
    def copy(self) -> ProbePositionsBuilder:
        pass

    @abstractmethod
    def build(self) -> ProbePositionSequence:
        pass


class FromMemoryProbePositionsBuilder(ProbePositionsBuilder):
    def __init__(self, settings: ProbePositionsSettings, points: Sequence[ProbePosition]) -> None:
        super().__init__(settings, 'from_memory')
        self._settings = settings
        self._scan = ProbePositionSequence(points)

    def copy(self) -> FromMemoryProbePositionsBuilder:
        return FromMemoryProbePositionsBuilder(self._settings, self._scan)

    def build(self) -> ProbePositionSequence:
        return self._scan


class FromFileProbePositionsBuilder(ProbePositionsBuilder):
    def __init__(
        self,
        settings: ProbePositionsSettings,
        file_path: Path,
        file_type: str,
        file_reader: ProbePositionFileReader,
    ) -> None:
        super().__init__(settings, 'from_file')
        self._settings = settings
        self.file_path = settings.file_path.copy()
        self.file_path.set_value(file_path)
        self._add_parameter('file_path', self.file_path)
        self.file_type = settings.file_type.copy()
        self.file_type.set_value(file_type)
        self._add_parameter('file_type', self.file_type)
        self._file_reader = file_reader

    def copy(self) -> FromFileProbePositionsBuilder:
        return FromFileProbePositionsBuilder(
            self._settings,
            self.file_path.get_value(),
            self.file_type.get_value(),
            self._file_reader,
        )

    def build(self) -> ProbePositionSequence:
        file_path = self.file_path.get_value()
        file_type = self.file_type.get_value()
        logger.debug(f'Reading "{file_path}" as "{file_type}"')

        try:
            scan = self._file_reader.read(file_path)
        except Exception as exc:
            raise RuntimeError(f'Failed to read "{file_path}"') from exc

        return scan



from __future__ import annotations
from typing import TypeAlias
import logging

import numpy
import numpy.typing

from ptychodus.api.common import ComplexArrayType
from ptychodus.api.geometry import ImageExtent

Float32ArrayType: TypeAlias = numpy.typing.NDArray[numpy.float32]

logger = logging.getLogger(__name__)


class PatternCircularBuffer:
    def __init__(self, extent: ImageExtent, max_size: int) -> None:
        self._buffer: Float32ArrayType = numpy.zeros(
            (max_size, *extent.get_shape()),
            dtype=numpy.float32,
        )
        self._pos = 0
        self._full = False

    @classmethod
    def create_zero_sized(cls) -> PatternCircularBuffer:
        return cls(ImageExtent(0, 0), 0)

    @property
    def is_zero_sized(self) -> bool:
        return self._buffer.size == 0

    def append(self, array: Float32ArrayType) -> None:
        self._buffer[self._pos, :, :] = array
        self._pos += 1

        if self._pos == self._buffer.shape[0]:
            self._pos = 0
            self._full = True

    def get_buffer(self) -> Float32ArrayType:
        return self._buffer if self._full else self._buffer[: self._pos]

    def set_buffer(self, array: Float32ArrayType) -> None:
        self._buffer = array
        self._pos = 0
        self._full = True


class ObjectPatchCircularBuffer:
    def __init__(self, extent: ImageExtent, channels: int, max_size: int) -> None:
        self._buffer: Float32ArrayType = numpy.zeros(
            (max_size, channels, *extent.get_shape()),
            dtype=numpy.float32,
        )
        self._pos = 0
        self._full = False

    @classmethod
    def create_zero_sized(cls) -> ObjectPatchCircularBuffer:
        return cls(ImageExtent(0, 0), 0, 0)

    @property
    def is_zero_sized(self) -> bool:
        return self._buffer.size == 0

    def append(self, array: ComplexArrayType) -> None:
        self._buffer[self._pos, 0, :, :] = numpy.angle(array).astype(numpy.float32)

        if self._buffer.shape[1] > 1:
            self._buffer[self._pos, 1, :, :] = numpy.absolute(array).astype(numpy.float32)

        self._pos += 1

        if self._pos == self._buffer.shape[0]:
            self._pos = 0
            self._full = True

    def get_buffer(self) -> Float32ArrayType:
        return self._buffer if self._full else self._buffer[: self._pos]

    def set_buffer(self, array: Float32ArrayType) -> None:
        self._buffer = array
        self._pos = 0
        self._full = True



from __future__ import annotations
from collections.abc import Iterator
from dataclasses import dataclass
from pathlib import Path
from types import TracebackType
from typing import overload
import logging
import queue
import threading
import time

from ptychodus.api.io import save_product
from ptychodus.api.observer import Observable
from ptychodus.api.product import Product
from ptychodus.api.reconstructor import (
    ReconstructInput,
    ReconstructOutput,
    Reconstructor,
    TrainableReconstructor,
)

from ..product import ProductRepositoryItem
from ..task_manager import ForegroundTaskManager

__all__ = [
    'ProcessingProgressMonitor',
    'ProcessingContext',
    'ReconstructBackgroundTask',
    'TrainBackgroundTask',
]

logger = logging.getLogger(__name__)


class NotifyObserversTask:
    def __init__(self, observable: Observable) -> None:
        self._observable = observable

    def __call__(self) -> None:
        self._observable.notify_observers()


class UpdateProductTask:
    def __init__(self, product_item: ProductRepositoryItem, product: Product) -> None:
        self._product_item = product_item
        self._product = product

    def __call__(self) -> None:
        name = self._product_item.get_name()
        self._product_item.assign(self._product)
        self._product_item.set_name(name)


class ProcessingLogHandler(logging.Handler):
    def __init__(self) -> None:
        super().__init__()
        self._log: queue.Queue[str] = queue.Queue()

    def messages(self) -> Iterator[str]:
        while True:
            try:
                yield self._log.get(block=False)
                self._log.task_done()
            except queue.Empty:
                break

    def emit(self, record: logging.LogRecord) -> None:
        text = self.format(record)
        self._log.put(text)


class ProcessingProgressMonitor(Observable):
    def __init__(
        self,
        foreground_task_manager: ForegroundTaskManager,
        processing_event: threading.Event,
        stop_event: threading.Event,
    ) -> None:
        super().__init__()
        self._foreground_task_manager = foreground_task_manager
        self._processing_event = processing_event
        self._stop_event = stop_event

        self._log_handler = ProcessingLogHandler()
        self._log_handler.setFormatter(
            logging.Formatter('%(asctime)s [%(levelname)s] %(name)s: %(message)s')
        )
        self._progress_lock = threading.Lock()
        self._progress_goal = 0
        self._progress = 0

    def get_log_handler(self) -> ProcessingLogHandler:
        return self._log_handler

    @property
    def is_processing(self) -> bool:
        return self._processing_event.is_set()

    def stop_processing(self) -> None:
        self._stop_event.set()

    def get_progress(self) -> int:
        with self._progress_lock:
            return self._progress

    def get_progress_goal(self) -> int:
        with self._progress_lock:
            return self._progress_goal

    def _notify_observers_foreground(self) -> None:
        task = NotifyObserversTask(self)
        self._foreground_task_manager.put_foreground_task(task)

    def _update_progress(self, progress: int, progress_goal: int) -> None:
        with self._progress_lock:
            is_changed = False

            if self._progress != progress:
                self._progress = progress
                is_changed = True

            if self._progress_goal != progress_goal:
                self._progress_goal = progress_goal
                is_changed = True

            if is_changed:
                self._notify_observers_foreground()


class ProcessingContext:
    def __init__(self, foreground_task_manager: ForegroundTaskManager) -> None:
        self._foreground_task_manager = foreground_task_manager
        self._processing_event = threading.Event()
        self._stop_event = threading.Event()
        self._progress_monitor = ProcessingProgressMonitor(
            foreground_task_manager, self._processing_event, self._stop_event
        )

    def get_progress_monitor(self) -> ProcessingProgressMonitor:
        return self._progress_monitor

    def get_log_handler(self) -> ProcessingLogHandler:
        return self._progress_monitor.get_log_handler()

    def update_progress(self, progress: int, progress_goal: int) -> None:
        self._progress_monitor._update_progress(progress, progress_goal)

    def update_product(
        self, product_item: ProductRepositoryItem, result: ReconstructOutput
    ) -> None:
        # TODO log handler messages to product comments
        task = UpdateProductTask(product_item, result.product)
        self._foreground_task_manager.put_foreground_task(task)

    @property
    def is_stopping(self) -> bool:
        return self._stop_event.is_set()

    def __enter__(self) -> ProcessingContext:
        self._processing_event.set()
        self._stop_event.clear()
        self._progress_monitor._notify_observers_foreground()
        return self

    @overload
    def __exit__(self, exception_type: None, exception_value: None, traceback: None) -> None: ...

    @overload
    def __exit__(
        self,
        exception_type: type[BaseException],
        exception_value: BaseException,
        traceback: TracebackType,
    ) -> None: ...

    def __exit__(
        self,
        exception_type: type[BaseException] | None,
        exception_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> None:
        self._processing_event.clear()
        self._progress_monitor._notify_observers_foreground()


@dataclass(frozen=True)
class ReconstructBackgroundTask:
    context: ProcessingContext
    reconstructor: Reconstructor
    reconstruct_input: ReconstructInput
    product_item: ProductRepositoryItem
    output_product_file: Path | None
    finished_event: threading.Event

    def __call__(self) -> None:
        with self.context as context:
            context.update_progress(0, self.reconstructor.get_progress_goal())
            tic = time.perf_counter()

            for result in self.reconstructor.reconstruct(self.reconstruct_input):
                context.update_progress(result.progress, self.reconstructor.get_progress_goal())
                context.update_product(self.product_item, result)
                of = self.output_product_file

                if of is not None:
                    of = of.parent / f'{of.stem}.{result.progress:06d}.{of.suffix}'
                    save_product(of, result.product)

                if context.is_stopping:
                    break

            toc = time.perf_counter()
            logger.info(f'Reconstruction time {toc - tic:.4f} seconds.')

        self.finished_event.set()


@dataclass(frozen=True)
class TrainBackgroundTask:
    context: ProcessingContext
    trainer: TrainableReconstructor
    reconstruct_input: ReconstructInput
    input_path: Path
    output_path: Path
    finished_event: threading.Event

    def __call__(self) -> None:
        with self.context as context:
            context.update_progress(0, self.trainer.get_progress_goal())
            tic = time.perf_counter()

            for result in self.trainer.train(self.input_path, self.output_path):
                context.update_progress(result.progress, self.trainer.get_progress_goal())

                if context.is_stopping:
                    break

            toc = time.perf_counter()
            logger.info(f'Training time {toc - tic:.4f} seconds.')

        self.finished_event.set()



"""Fluorescence data structures and enhancement plugin interfaces."""

from abc import ABC, abstractmethod
from collections.abc import Sequence
from dataclasses import dataclass
from pathlib import Path

from .common import RealArrayType
from .product import Product


@dataclass(frozen=True)
class ElementMap:
    """2D spatial map of fluorescence signal for a single element in counts per second."""

    name: str
    counts_per_second: RealArrayType


@dataclass(frozen=True)
class FluorescenceDataset:
    """Collection of element maps with metadata."""

    element_maps: Sequence[ElementMap]
    counts_per_second_path: str
    channel_names_path: str

    # TODO need to communicate association between element map pixels and scan order.
    #      integer-valued, same shape as counts_per_second
    # scan_indexes: IntegerArray


class FluorescenceEnhancingAlgorithm(ABC):
    """Algorithm that uses a ptychography data product to enhance a fluorescence dataset."""

    @abstractmethod
    def enhance(self, dataset: FluorescenceDataset, product: Product) -> FluorescenceDataset:
        pass


class FluorescenceFileReader(ABC):
    """Plugin interface for reading fluorescence datasets."""

    @abstractmethod
    def read(self, file_path: Path) -> FluorescenceDataset:
        """Read a fluorescence dataset from file."""
        pass


class FluorescenceFileWriter(ABC):
    """Plugin interface for writing fluorescence datasets."""

    @abstractmethod
    def write(self, file_path: Path, dataset: FluorescenceDataset) -> None:
        """Write a fluorescence dataset to file."""
        pass


class UpscalingStrategy(ABC):
    """Remaps element concentrations from the regular scan grid to the upscaled grid using ptychography-corrected probe positions."""

    @abstractmethod
    def __call__(self, emap: ElementMap, product: Product) -> ElementMap:
        pass


class DeconvolutionStrategy(ABC):
    """Deconvolves the probe kernel from the accumulated array to enhance the element map."""

    @abstractmethod
    def __call__(self, emap: ElementMap, product: Product) -> ElementMap:
        pass



from __future__ import annotations
from typing import Any

from PyQt5.QtCore import Qt, QAbstractListModel, QModelIndex, QObject
from PyQt5.QtGui import QFont
from PyQt5.QtWidgets import (
    QHBoxLayout,
    QListView,
    QPushButton,
    QVBoxLayout,
    QWidget,
)

from ptychodus.api.observer import Observable, Observer, SequenceObserver

from ..model.automation import (
    AutomationDataset,
    AutomationDatasetRepository,
    AutomationDatasetState,
    AutomationPresenter,
    AutomationSettings,
)
from .parametric import ParameterViewBuilder, ParameterViewController
from .data import FileDialogFactory


class AutomationDatasetListModel(QAbstractListModel):
    def __init__(
        self, repository: AutomationDatasetRepository, parent: QObject | None = None
    ) -> None:
        super().__init__(parent)
        self._repository = repository

    def data(self, index: QModelIndex, role: int = Qt.ItemDataRole.DisplayRole) -> Any:
        if index.isValid():
            dataset = self._repository[index.row()]

            match role:
                case Qt.ItemDataRole.DisplayRole:
                    return dataset.label
                case Qt.ItemDataRole.FontRole:
                    font = QFont()

                    match dataset.state:
                        case AutomationDatasetState.WAITING:
                            font.setItalic(True)
                        case AutomationDatasetState.RUNNING:
                            font.setBold(True)

                    return font

    def rowCount(self, parent: QModelIndex = QModelIndex()) -> int:  # noqa: N802
        return len(self._repository)


class AutomationDatasetListViewController(
    ParameterViewController, SequenceObserver[AutomationDataset]
):
    def __init__(self, repository: AutomationDatasetRepository) -> None:
        super().__init__()
        self._model = AutomationDatasetListModel(repository)
        self._widget = QListView()
        self._widget.setModel(self._model)

        repository.add_observer(self)

    def get_widget(self) -> QWidget:
        return self._widget

    def handle_item_inserted(self, index: int, item: AutomationDataset) -> None:
        self._model.beginInsertRows(QModelIndex(), index, index)
        self._model.endInsertRows()

    def handle_item_changed(self, index: int, item: AutomationDataset) -> None:
        top_left = self._model.index(index, 0)
        bottom_right = self._model.index(index, self._model.columnCount())
        self._model.dataChanged.emit(top_left, bottom_right)

    def handle_item_removed(self, index: int, item: AutomationDataset) -> None:
        self._model.beginRemoveRows(QModelIndex(), index, index)
        self._model.endRemoveRows()


class AutomationButtonsViewController(ParameterViewController, Observer):
    def __init__(
        self, repository: AutomationDatasetRepository, presenter: AutomationPresenter
    ) -> None:
        super().__init__()
        self._repository = repository
        self._presenter = presenter
        self._load_button = QPushButton('Load')
        self._watch_button = QPushButton('Watch')
        self._clear_button = QPushButton('Clear')
        self._widget = QWidget()

        self._load_button.clicked.connect(presenter.load_existing_datasets_to_repository)
        self._watch_button.setCheckable(True)
        self._watch_button.toggled.connect(presenter.set_watcher_enabled)
        self._clear_button.clicked.connect(repository.clear)

        layout = QHBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(self._load_button)
        layout.addWidget(self._watch_button)
        layout.addWidget(self._clear_button)
        self._widget.setLayout(layout)

        self._sync_model_to_view()
        presenter.add_observer(self)

    def get_widget(self) -> QWidget:
        return self._widget

    def _sync_model_to_view(self) -> None:
        self._watch_button.setChecked(self._presenter.is_watcher_enabled())

    def _update(self, observable: Observable) -> None:
        if observable is self._presenter:
            self._sync_model_to_view()


class AutomationController:
    def __init__(
        self,
        settings: AutomationSettings,
        repository: AutomationDatasetRepository,
        presenter: AutomationPresenter,
        view: QWidget,
        file_dialog_factory: FileDialogFactory,
    ) -> None:
        super().__init__()
        self._settings = settings
        self._repository = repository
        self._presenter = presenter
        self._view = view
        self._file_dialog_factory = file_dialog_factory

        view_builder = ParameterViewBuilder(file_dialog_factory)

        processing_group = 'Processing'
        view_builder.add_combo_box(
            settings.workflow, presenter.available_workflows(), 'Workflow:', group=processing_group
        )
        view_builder.add_directory_chooser(
            settings.data_directory, 'Data Directory:', group=processing_group
        )

        watchdog_group = 'Watchdog'
        view_builder.add_spin_box(settings.watchdog_delay_s, 'Delay [sec]:', group=watchdog_group)
        view_builder.add_check_box(
            settings.use_watchdog_polling_observer, 'Use Polling Observer', group=watchdog_group
        )

        list_view_controller = AutomationDatasetListViewController(repository)
        view_builder.add_view_controller_to_bottom(list_view_controller)

        buttons_view_controller = AutomationButtonsViewController(repository, presenter)
        view_builder.add_view_controller_to_bottom(buttons_view_controller)

        contents = view_builder.build_widget()

        layout = QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(contents)
        view.setLayout(layout)



from PyQt5.QtWidgets import QVBoxLayout, QWidget

from ...model.ptychi import (
    PtyChiAutodiffSettings,
    PtyChiDMSettings,
    PtyChiLSQMLSettings,
    PtyChiPIESettings,
    PtyChiReconstructorLibrary,
)

from ..processing import ReconstructorViewControllerFactory
from .object import PtyChiObjectViewController
from .opr import PtyChiOPRViewController
from .probe_positions import PtyChiProbePositionsViewController
from .probe import PtyChiProbeViewController
from .reconstructor import PtyChiReconstructorViewController

__all__ = ['PtyChiViewControllerFactory']


class PtyChiViewController(QWidget):
    def __init__(
        self,
        model: PtyChiReconstructorLibrary,
        reconstructor_name: str,
        parent: QWidget | None = None,
    ) -> None:
        super().__init__(parent)
        autodiff_settings: PtyChiAutodiffSettings | None = None
        dm_settings: PtyChiDMSettings | None = None
        lsqml_settings: PtyChiLSQMLSettings | None = None
        pie_settings: PtyChiPIESettings | None = None

        match reconstructor_name:
            case 'Autodiff':
                autodiff_settings = model.autodiff_settings
            case 'DM':
                dm_settings = model.dm_settings
            case 'LSQML':
                lsqml_settings = model.lsqml_settings
            case 'PIE' | 'ePIE' | 'rPIE':
                pie_settings = model.pie_settings

        self._reconstructor_view_controller = PtyChiReconstructorViewController(
            model.settings,
            autodiff_settings,
            dm_settings,
            lsqml_settings,
            model.enumerators,
            model.device_repository,
        )
        self._object_view_controller = PtyChiObjectViewController(
            model.object_settings,
            dm_settings,
            lsqml_settings,
            pie_settings,
            model.settings.num_epochs,
            model.enumerators,
        )
        self._probe_view_controller = PtyChiProbeViewController(
            model.probe_settings,
            dm_settings,
            lsqml_settings,
            pie_settings,
            model.settings.num_epochs,
            model.enumerators,
        )
        self._probe_positions_view_controller = PtyChiProbePositionsViewController(
            model.probe_position_settings,
            model.settings.num_epochs,
            model.enumerators,
        )
        self._opr_view_controller = PtyChiOPRViewController(
            model.opr_settings, model.settings.num_epochs, model.enumerators
        )

        layout = QVBoxLayout()
        layout.addWidget(self._reconstructor_view_controller.get_widget())
        layout.addWidget(self._object_view_controller.get_widget())
        layout.addWidget(self._probe_view_controller.get_widget())
        layout.addWidget(self._probe_positions_view_controller.get_widget())
        layout.addWidget(self._opr_view_controller.get_widget())
        layout.addStretch()
        self.setLayout(layout)


class PtyChiViewControllerFactory(ReconstructorViewControllerFactory):
    def __init__(self, model: PtyChiReconstructorLibrary) -> None:
        super().__init__()
        self._model = model

    @property
    def name(self) -> str:
        return 'pty-chi'

    def create_view_controller(self, reconstructor_name: str) -> QWidget:
        return PtyChiViewController(self._model, reconstructor_name)



from dataclasses import dataclass
from pathlib import Path
import csv
import logging
import re

from ptychodus.api.plugins import PluginRegistry
from ptychodus.api.workflow import FileBasedWorkflow, WorkflowAPI

logger = logging.getLogger(__name__)


class PtychodusAutoloadProductFileBasedWorkflow(FileBasedWorkflow):
    @property
    def is_watch_recursive(self) -> bool:
        return True

    def get_watch_file_pattern(self) -> str:
        return 'product-out.h5'

    def execute(self, api: WorkflowAPI, file_path: Path) -> None:
        api.load_product(file_path)


class APS2IDFileBasedWorkflow(FileBasedWorkflow):
    @property
    def is_watch_recursive(self) -> bool:
        return False

    def get_watch_file_pattern(self) -> str:
        return '*.csv'

    def execute(self, api: WorkflowAPI, file_path: Path) -> None:
        scan_name = file_path.stem
        scan_id = int(re.findall(r'\d+', scan_name)[-1])

        diffraction_file_path = file_path.parents[1] / 'raw_data' / f'scan{scan_id}_master.h5'
        api.load_diffraction_data(diffraction_file_path)
        product_api = api.create_product(f'scan{scan_id}')
        product_api.load_probe_positions(file_path)
        product_api.generate_probe()
        product_api.generate_object()
        product_api.reconstruct_remote()


class APS26IDFileBasedWorkflow(FileBasedWorkflow):
    @property
    def is_watch_recursive(self) -> bool:
        return False

    def get_watch_file_pattern(self) -> str:
        return '*.mda'

    def execute(self, api: WorkflowAPI, file_path: Path) -> None:
        scan_name = file_path.stem
        scan_id = int(re.findall(r'\d+', scan_name)[-1])

        diffraction_dir_path = file_path.parents[1] / 'h5'

        for diffraction_file_path in diffraction_dir_path.glob(f'scan_{scan_id}_*.h5'):
            digits = int(re.findall(r'\d+', diffraction_file_path.stem)[-1])

            if digits != 0:
                break

        api.load_diffraction_data(diffraction_file_path)
        product_api = api.create_product(f'scan_{scan_id}')
        product_api.load_probe_positions(file_path)
        product_api.generate_probe()
        product_api.generate_object()
        product_api.reconstruct_remote()


@dataclass(frozen=True)
class APS31IDEMetadata:
    scan_no: int
    golden_angle: str
    encoder_angle: str
    measurement_id: str
    subtomo_no: str
    detector_position: str
    label: str

    def __str__(self) -> str:
        return f"""scan_no={self.scan_no}
        golden_angle={self.golden_angle}
        encoder_angle={self.encoder_angle}
        measurement_id={self.measurement_id}
        subtomo_no={self.subtomo_no}
        detector_position={self.detector_position}
        label={self.label}
        """


class APS31IDEFileBasedWorkflow(FileBasedWorkflow):
    @property
    def is_watch_recursive(self) -> bool:
        return True

    def get_watch_file_pattern(self) -> str:
        return '*.h5'

    def execute(self, api: WorkflowAPI, file_path: Path) -> None:
        if file_path.parent.name != 'eiger_4':
            return

        experiment_dir = file_path.parents[3]
        scan_num = int(re.findall(r'\d+', file_path.stem)[0])
        scan_file = experiment_dir / 'scan_positions' / f'scan_{scan_num:05d}.dat'
        scan_numbers_file = experiment_dir / 'dat-files' / 'tomography_scannumbers.txt'
        metadata: APS31IDEMetadata | None = None

        with scan_numbers_file.open(newline='') as csv_file:
            csv_reader = csv.reader(csv_file, delimiter=' ')

            for row in csv_reader:
                if row[0].startswith('#'):
                    continue

                if len(row) != 7:
                    logger.warning('Unexpected row in tomography_scannumbers.txt!')
                    logger.debug(row)
                    continue

                try:
                    row_no = int(row[0])
                except ValueError:
                    logger.warning('Failed to parse row ID in tomography_scannumbers.txt!')
                    logger.debug(row[0])
                else:
                    if row_no == scan_num:
                        metadata = APS31IDEMetadata(
                            scan_no=scan_num,
                            golden_angle=str(row[1]),
                            encoder_angle=str(row[2]),
                            measurement_id=str(row[3]),
                            subtomo_no=str(row[4]),
                            detector_position=str(row[5]),
                            label=str(row[6]),
                        )
                        break

        if metadata is None:
            logger.warning(f'Failed to locate metadata for {scan_num}!')
        else:
            product_name = f'scan{scan_num:05d}_' + metadata.label
            api.load_diffraction_data(file_path)
            input_product_api = api.create_product(product_name, comments=str(metadata))
            input_product_api.load_probe_positions(scan_file)
            input_product_api.generate_probe()
            input_product_api.generate_object()
            # TODO would prefer to write instructions and submit to queue
            output_product_file = experiment_dir / 'ptychodus' / f'{product_name}.h5'
            input_product_api.reconstruct_local(output_product_file=output_product_file)


def register_plugins(registry: PluginRegistry) -> None:
    registry.file_based_workflows.register_plugin(
        PtychodusAutoloadProductFileBasedWorkflow(),
        simple_name='Autoload_Product',
        display_name='Autoload Product',
    )
    registry.file_based_workflows.register_plugin(
        APS2IDFileBasedWorkflow(),
        simple_name='APS_2ID',
        display_name='APS 2-ID',
    )
    registry.file_based_workflows.register_plugin(
        APS26IDFileBasedWorkflow(),
        simple_name='APS_26ID',
        display_name='APS 26-ID',
    )
    registry.file_based_workflows.register_plugin(
        APS31IDEFileBasedWorkflow(),
        simple_name='APS_31IDE',
        display_name='APS 31-ID-E',
    )



from __future__ import annotations
from collections.abc import Iterator
import logging

from ptychodus.api.reconstructor import (
    NullReconstructor,
    Reconstructor,
    ReconstructorLibrary,
    TrainableReconstructor,
)
from ptychodus.api.settings import SettingsRegistry

from .settings import PtychoNNModelSettings, PtychoNNTrainingSettings

logger = logging.getLogger(__name__)


class PtychoNNReconstructorLibrary(ReconstructorLibrary):
    def __init__(
        self, settings_registry: SettingsRegistry, is_developer_mode_enabled: bool
    ) -> None:
        super().__init__('ptychonn')
        self.model_settings = PtychoNNModelSettings(settings_registry)
        self.training_settings = PtychoNNTrainingSettings(settings_registry)
        self._reconstructors: list[TrainableReconstructor] = list()

        try:
            from .model import PtychoNNModelProvider
            from .reconstructor import PtychoNNTrainableReconstructor
        except ModuleNotFoundError:
            logger.info('PtychoNN not found.')

            if is_developer_mode_enabled:
                self._reconstructors.append(NullReconstructor('PhaseOnly'))
                self._reconstructors.append(NullReconstructor('AmplitudePhase'))
        else:
            phase_only_model_provider = PtychoNNModelProvider(
                self.model_settings, self.training_settings, enable_amplitude=False
            )
            amplitude_phase_model_provider = PtychoNNModelProvider(
                self.model_settings, self.training_settings, enable_amplitude=True
            )

            self._reconstructors.append(
                PtychoNNTrainableReconstructor(
                    self.model_settings, self.training_settings, phase_only_model_provider
                )
            )
            self._reconstructors.append(
                PtychoNNTrainableReconstructor(
                    self.model_settings, self.training_settings, amplitude_phase_model_provider
                )
            )

    @property
    def name(self) -> str:
        return 'PtychoNN'

    def __iter__(self) -> Iterator[Reconstructor]:
        return iter(self._reconstructors)



from __future__ import annotations
from collections.abc import Sequence
from dataclasses import dataclass
from pathlib import Path
from typing import Any
import logging

import numpy

from ptychodus.api.common import ComplexArrayType, RealArrayType
from ptychodus.api.geometry import PixelGeometry
from ptychodus.api.object import ObjectCenter
from ptychodus.api.observer import Observable

from ..product import ProductRepository

__all__ = [
    'XMCDAnalyzer',
    'XMCDResult',
]

logger = logging.getLogger(__name__)


def robust_log_amplitude(array: ComplexArrayType, epsilon: float = 1e-10) -> RealArrayType:
    return numpy.log(numpy.maximum(numpy.absolute(array), epsilon))


@dataclass(frozen=True)
class XMCDResult:
    polar_difference: RealArrayType
    polar_sum: RealArrayType
    polar_ratio: RealArrayType
    pixel_geometry: PixelGeometry
    center: ObjectCenter


class XMCDAnalyzer(Observable):
    # TODO feature request: want ability to align/add reconstructed slices
    #      of repeat scans for each polarization separately to improve statistics

    def __init__(self, repository: ProductRepository) -> None:
        super().__init__()
        self._repository = repository

        self._lcirc_product_index = -1
        self._rcirc_product_index = -1
        self._result: XMCDResult | None = None

    def set_lcirc_product(self, lcirc_product_index: int) -> None:
        if self._lcirc_product_index != lcirc_product_index:
            self._lcirc_product_index = lcirc_product_index
            self._lcirc_product_data = None
            self.notify_observers()

    def get_lcirc_product(self) -> int:
        return self._lcirc_product_index

    def get_lcirc_product_name(self) -> str:
        lcirc_product = self._repository[self._lcirc_product_index]
        return lcirc_product.get_name()

    def set_rcirc_product(self, rcirc_product_index: int) -> None:
        if self._rcirc_product_index != rcirc_product_index:
            self._rcirc_product_index = rcirc_product_index
            self._rcirc_product_data = None
            self.notify_observers()

    def get_rcirc_product(self) -> int:
        return self._rcirc_product_index

    def get_rcirc_product_name(self) -> str:
        rcirc_product = self._repository[self._rcirc_product_index]
        return rcirc_product.get_name()

    def analyze(self) -> None:
        lcirc_object = self._repository[self._lcirc_product_index].get_object_item().get_object()
        rcirc_object = self._repository[self._rcirc_product_index].get_object_item().get_object()

        lcirc_object_geometry = lcirc_object.get_geometry()
        rcirc_object_geometry = rcirc_object.get_geometry()

        if lcirc_object_geometry.width_px != rcirc_object_geometry.width_px:
            raise ValueError('Object width mismatch!')

        if lcirc_object_geometry.height_px != rcirc_object_geometry.height_px:
            raise ValueError('Object height mismatch!')

        if lcirc_object_geometry.pixel_width_m != rcirc_object_geometry.pixel_width_m:
            raise ValueError('Object pixel width mismatch!')

        if lcirc_object_geometry.pixel_height_m != rcirc_object_geometry.pixel_height_m:
            raise ValueError('Object pixel height mismatch!')

        # TODO align lcirc_array/rcirc_array
        lcirc_log_amp = robust_log_amplitude(lcirc_object.get_layers_flattened())
        rcirc_log_amp = robust_log_amplitude(rcirc_object.get_layers_flattened())

        polar_difference = lcirc_log_amp - rcirc_log_amp
        polar_sum = lcirc_log_amp + rcirc_log_amp
        polar_ratio = numpy.divide(
            polar_difference,
            polar_sum,
            out=numpy.zeros_like(polar_sum),
            where=(numpy.absolute(polar_sum) > 1e-10),
        )

        self._result = XMCDResult(
            polar_difference=polar_difference,
            polar_sum=polar_sum,
            polar_ratio=polar_ratio,
            pixel_geometry=rcirc_object.get_pixel_geometry(),
            center=rcirc_object.get_center(),
        )
        self.notify_observers()

    def get_result(self) -> XMCDResult:
        if self._result is None:
            raise ValueError('No analyzed data!')

        return self._result

    def get_save_file_filters(self) -> Sequence[str]:
        return [self.get_save_file_filter()]

    def get_save_file_filter(self) -> str:
        return 'NumPy Zipped Archive (*.npz)'

    def save_data(self, file_path: Path) -> None:
        if self._result is None:
            raise ValueError('No analyzed data!')

        contents: dict[str, Any] = {
            'polar_difference': self._result.polar_difference,
            'polar_sum': self._result.polar_sum,
            'polar_ratio': self._result.polar_ratio,
            'pixel_height_m': self._result.pixel_geometry.height_m,
            'pixel_width_m': self._result.pixel_geometry.width_m,
            'center_x_m': self._result.center.coordinate_x_m,
            'center_y_m': self._result.center.coordinate_y_m,
        }

        numpy.savez_compressed(file_path, allow_pickle=False, **contents)



from PyQt5.QtCore import Qt
from PyQt5.QtGui import QIcon
from PyQt5.QtWidgets import (
    QFrame,
    QHBoxLayout,
    QListView,
    QPushButton,
    QScrollArea,
    QSizePolicy,
    QSplitter,
    QPlainTextEdit,
    QWidget,
)


class AgentView(QWidget):
    pass


class AgentInputView(QFrame):
    def __init__(self, parent: QWidget | None = None) -> None:
        super().__init__(parent)
        self.text_edit = QPlainTextEdit()

        send_button_size_policy = QSizePolicy(
            QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Preferred
        )
        self.send_button = QPushButton(QIcon(':/icons/send'), 'Send')
        self.send_button.setSizePolicy(send_button_size_policy)

        layout = QHBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(self.text_edit)
        layout.addWidget(self.send_button)
        self.setLayout(layout)


class AgentChatView(QSplitter):
    def __init__(self, parent: QWidget | None = None) -> None:
        super().__init__(Qt.Orientation.Vertical, parent)
        self.message_list_view = QListView()
        self.scroll_area = QScrollArea()
        self.scroll_area.setWidgetResizable(True)
        self.scroll_area.setWidget(self.message_list_view)
        self.input_view = AgentInputView()

        self.addWidget(self.scroll_area)
        self.addWidget(self.input_view)

        self.setStretchFactor(0, 2)
        self.setStretchFactor(1, 0)



from collections.abc import Mapping
from pathlib import Path
import logging
import re
import sys

from tifffile import TiffFile
import numpy

from ptychodus.api.geometry import ImageExtent
from ptychodus.api.diffraction import (
    DiffractionDataset,
    DiffractionFileReader,
    DiffractionMetadata,
    DiffractionArray,
    DiffractionPatterns,
    DiffractionIndexes,
    SimpleDiffractionDataset,
)
from ptychodus.api.plugins import PluginRegistry
from ptychodus.api.tree import SimpleTreeNode

logger = logging.getLogger(__name__)


class TiffDiffractionPatternArray(DiffractionArray):
    def __init__(self, file_path: Path, index: int) -> None:
        super().__init__()
        self._file_path = file_path
        self._indexes = numpy.array([index])

    def get_label(self) -> str:
        return self._file_path.stem

    def get_indexes(self) -> DiffractionIndexes:
        return self._indexes

    def get_patterns(self) -> DiffractionPatterns:
        with TiffFile(self._file_path) as tiff:
            data = tiff.asarray()

        if data.ndim == 2:
            data = data[numpy.newaxis, :, :]

        return data


class TiffDiffractionFileReader(DiffractionFileReader):
    def _get_file_series(self, file_path: Path) -> tuple[Mapping[int, Path], str]:
        file_path_dict: dict[int, Path] = dict()

        digits = re.findall(r'\d+', file_path.stem)
        longest_digits = max(digits, key=len)
        file_pattern = file_path.name.replace(longest_digits, f'(\\d{{{len(longest_digits)}}})')

        for fp in file_path.parent.iterdir():
            z = re.match(file_pattern, fp.name)

            if z:
                index = int(z.group(1).lstrip('0'))
                file_path_dict[index] = fp

        return file_path_dict, file_pattern

    def read(self, file_path: Path) -> DiffractionDataset:
        file_path_mapping, file_pattern = self._get_file_series(file_path)
        contents_tree = SimpleTreeNode.create_root(['Name', 'Type', 'Details'])
        array_list: list[DiffractionArray] = list()

        for idx, (_, fp) in enumerate(sorted(file_path_mapping.items())):  # TODO use keys
            array = TiffDiffractionPatternArray(fp, idx)
            contents_tree.create_child([array.get_label(), 'TIFF', str(idx)])
            array_list.append(array)

        with TiffFile(file_path) as tiff:
            data = tiff.asarray()

        if data.ndim == 2:
            data = data[numpy.newaxis, :, :]

        num_patterns_per_array, detector_height, detector_width = data.shape

        metadata = DiffractionMetadata(
            num_patterns_per_array=[num_patterns_per_array] * len(array_list),
            pattern_dtype=data.dtype,
            detector_extent=ImageExtent(detector_width, detector_height),
            file_path=file_path.parent / file_pattern,
        )

        return SimpleDiffractionDataset(metadata, contents_tree, array_list)


def register_plugins(registry: PluginRegistry) -> None:
    registry.diffraction_file_readers.register_plugin(
        TiffDiffractionFileReader(),
        simple_name='TIFF',
        display_name='Tagged Image File Format Files (*.tif *.tiff)',
    )


if __name__ == '__main__':
    file_path = Path(sys.argv[1])
    reader = TiffDiffractionFileReader()
    tiff_file = reader.read(file_path)
    print(tiff_file)



from abc import ABC, abstractmethod
from collections.abc import Sequence
from dataclasses import dataclass
from enum import Enum, auto
from typing import overload


class ChatRole(Enum):
    SYSTEM = auto()
    USER = auto()
    AGENT = auto()


@dataclass(frozen=True)
class ChatMessage:
    role: ChatRole
    content: str

    def to_dict(self) -> dict[str, str]:
        return {
            'role': self.role.name.lower(),
            'content': self.content,
        }

    def __str__(self) -> str:
        return str(self.to_dict())


class ChatTerminal(ABC):
    @abstractmethod
    def send_message(self, content: str) -> None:
        pass

    @abstractmethod
    def embed_texts(self, texts: Sequence[str]) -> Sequence[Sequence[float]]:
        pass


class ChatObserver(ABC):
    @abstractmethod
    def handle_new_message(self, message: ChatMessage, index: int) -> None:
        pass

    @abstractmethod
    def handle_chat_cleared(self) -> None:
        pass


class ChatHistory(Sequence[ChatMessage]):
    def __init__(self) -> None:
        self._messages: list[ChatMessage] = []
        self._observers: list[ChatObserver] = []

    @overload
    def __getitem__(self, index: int) -> ChatMessage: ...

    @overload
    def __getitem__(self, index: slice) -> Sequence[ChatMessage]: ...

    def __getitem__(self, index: int | slice) -> ChatMessage | Sequence[ChatMessage]:
        return self._messages[index]

    def __len__(self) -> int:
        return len(self._messages)

    def add_observer(self, observer: ChatObserver) -> None:
        if observer not in self._observers:
            self._observers.append(observer)

    def add_message(self, message: ChatMessage) -> None:
        index = len(self._messages)
        self._messages.append(message)

        for observer in self._observers:
            observer.handle_new_message(message, index)

    def clear(self) -> None:
        self._messages.clear()

        for observer in self._observers:
            observer.handle_chat_cleared()



from __future__ import annotations
import logging

from PyQt5.QtCore import QModelIndex, QSortFilterProxyModel, QStringListModel
from PyQt5.QtWidgets import QAbstractItemView, QDialog, QMessageBox

from ptychodus.api.observer import SequenceObserver

from ...model.product import ProbePositionsAPI, ProbePositionsRepository
from ...model.product.probe_positions import ProbePositionsRepositoryItem
from ...view.repository import RepositoryTableView
from ...view.probe_positions import ProbePositionsPlotView
from ...view.widgets import ComboBoxItemDelegate, ExceptionDialog
from ..data import FileDialogFactory
from ..helpers import (
    connect_current_changed_signal,
    connect_triggered_signal,
    create_brush_for_editable_cell,
)
from .editor_factory import ProbePositionsEditorViewControllerFactory
from .table_model import ProbePositionsTableModel

logger = logging.getLogger(__name__)


class ProbePositionsController(SequenceObserver[ProbePositionsRepositoryItem]):
    def __init__(
        self,
        repository: ProbePositionsRepository,
        api: ProbePositionsAPI,
        view: RepositoryTableView,
        plot_view: ProbePositionsPlotView,
        file_dialog_factory: FileDialogFactory,
        *,
        is_developer_mode_enabled: bool,
    ) -> None:
        super().__init__()
        self._repository = repository
        self._api = api
        self._view = view
        self._plot_view = plot_view
        self._file_dialog_factory = file_dialog_factory
        self._table_model = ProbePositionsTableModel(
            repository, api, create_brush_for_editable_cell(view)
        )
        self._table_proxy_model = QSortFilterProxyModel()
        self._editor_factory = ProbePositionsEditorViewControllerFactory()

        self._table_proxy_model.setSourceModel(self._table_model)
        self._table_proxy_model.dataChanged.connect(
            lambda top_left, bottom_right, roles: self._redraw_plot()
        )
        repository.add_observer(self)

        builder_list_model = QStringListModel()
        builder_list_model.setStringList([name for name in api.builder_names()])
        builder_item_delegate = ComboBoxItemDelegate(builder_list_model, view.table_view)

        view.table_view.setModel(self._table_proxy_model)
        view.table_view.setSortingEnabled(True)
        view.table_view.setSelectionBehavior(QAbstractItemView.SelectionBehavior.SelectRows)
        view.table_view.setItemDelegateForColumn(2, builder_item_delegate)
        connect_current_changed_signal(view.table_view, self._update_view)
        self._update_view(QModelIndex(), QModelIndex())

        view.table_view.horizontalHeader().sectionClicked.connect(
            lambda logical_index: self._redraw_plot()
        )

        load_from_file_action = view.button_box.load_menu.addAction('Open File...')
        connect_triggered_signal(load_from_file_action, self._load_current_scan_from_file)

        copy_action = view.button_box.load_menu.addAction('Copy...')
        connect_triggered_signal(copy_action, self._copy_to_current_scan)

        save_to_file_action = view.button_box.save_menu.addAction('Save File...')
        connect_triggered_signal(save_to_file_action, self._save_current_scan_to_file)

        sync_to_settings_action = view.button_box.save_menu.addAction('Sync To Settings')
        connect_triggered_signal(sync_to_settings_action, self._sync_current_scan_to_settings)

        view.copier_dialog.setWindowTitle('Copy Scan')
        view.copier_dialog.source_combo_box.setModel(self._table_model)
        view.copier_dialog.destination_combo_box.setModel(self._table_model)
        view.copier_dialog.finished.connect(self._finish_copying_scan)

        view.button_box.edit_button.clicked.connect(self._edit_current_scan)

        estimate_transform_action = view.button_box.analyze_menu.addAction('Estimate Transform...')
        connect_triggered_signal(estimate_transform_action, self._estimate_transform)
        estimate_transform_action.setEnabled(is_developer_mode_enabled)

    def _get_current_item_index(self) -> int:
        proxy_index = self._view.table_view.currentIndex()

        if proxy_index.isValid():
            model_index = self._table_proxy_model.mapToSource(proxy_index)
            return model_index.row()

        logger.warning('No current index!')
        return -1

    def _load_current_scan_from_file(self) -> None:
        item_index = self._get_current_item_index()

        if item_index < 0:
            return

        file_path, name_filter = self._file_dialog_factory.get_open_file_path(
            self._view,
            'Open Scan',
            name_filters=[nf for nf in self._api.get_open_file_filters()],
            selected_name_filter=self._api.get_open_file_filter(),
        )

        if file_path:
            try:
                self._api.open_probe_positions(item_index, file_path, file_type=name_filter)
            except Exception as err:
                logger.exception(err)
                ExceptionDialog.show_exception('File Reader', err)

    def _copy_to_current_scan(self) -> None:
        item_index = self._get_current_item_index()

        if item_index >= 0:
            self._view.copier_dialog.destination_combo_box.setCurrentIndex(item_index)
            self._view.copier_dialog.open()

    def _finish_copying_scan(self, result: int) -> None:
        if result == QDialog.DialogCode.Accepted:
            source_index = self._view.copier_dialog.source_combo_box.currentIndex()
            destination_index = self._view.copier_dialog.destination_combo_box.currentIndex()
            self._api.copy_scan(source_index, destination_index)

    def _edit_current_scan(self) -> None:
        item_index = self._get_current_item_index()

        if item_index < 0:
            return

        item_name = self._repository.get_name(item_index)
        item = self._repository[item_index]
        dialog = self._editor_factory.create_editor_dialog(item_name, item, self._view)
        dialog.open()

    def _save_current_scan_to_file(self) -> None:
        item_index = self._get_current_item_index()

        if item_index < 0:
            return

        file_path, name_filter = self._file_dialog_factory.get_save_file_path(
            self._view,
            'Save Scan',
            name_filters=[nameFilter for nameFilter in self._api.get_save_file_filters()],
            selected_name_filter=self._api.get_save_file_filter(),
        )

        if file_path:
            try:
                self._api.save_scan(item_index, file_path, name_filter)
            except Exception as err:
                logger.exception(err)
                ExceptionDialog.show_exception('File Writer', err)

    def _sync_current_scan_to_settings(self) -> None:
        item_index = self._get_current_item_index()

        if item_index < 0:
            logger.warning('No current item!')
        else:
            item = self._repository[item_index]
            item.sync_to_settings()

    def _redraw_plot(self) -> None:
        self._plot_view.axes.clear()

        for row in range(self._table_proxy_model.rowCount()):
            proxy_index = self._table_proxy_model.index(row, 0)
            item_index = self._table_proxy_model.mapToSource(proxy_index).row()

            if self._table_model.is_item_checked(item_index):
                item_name = self._repository.get_name(item_index)
                probe_positions = self._repository[item_index].get_probe_positions()
                x = [point.coordinate_x_m for point in probe_positions]
                y = [point.coordinate_y_m for point in probe_positions]
                self._plot_view.axes.plot(x, y, '.-', label=item_name, linewidth=1.5)

        self._plot_view.axes.invert_yaxis()
        self._plot_view.axes.axis('equal')
        self._plot_view.axes.grid(True)
        self._plot_view.axes.set_xlabel('X [m]')
        self._plot_view.axes.set_ylabel('Y [m]')

        if len(self._plot_view.axes.lines) > 0:
            self._plot_view.axes.legend(loc='best')

        self._plot_view.figure_canvas.draw()

    def _estimate_transform(self) -> None:  # TODO
        item_index = self._get_current_item_index()

        if item_index < 0:
            logger.warning('No current item!')
            return

        _ = QMessageBox.information(
            self._view,
            'Not Implemented',
            'Affine transform estimator is not yet implemented.',
        )

    def _update_view(self, current: QModelIndex, previous: QModelIndex) -> None:
        enabled = current.isValid()
        self._view.button_box.load_button.setEnabled(enabled)
        self._view.button_box.save_button.setEnabled(enabled)
        self._view.button_box.edit_button.setEnabled(enabled)
        self._view.button_box.analyze_button.setEnabled(enabled)
        self._redraw_plot()

    def handle_item_inserted(self, index: int, item: ProbePositionsRepositoryItem) -> None:
        self._table_model.insert_item(index, item)

    def handle_item_changed(self, index: int, item: ProbePositionsRepositoryItem) -> None:
        self._table_model.update_item(index, item)

        if self._table_model.is_item_checked(index):
            self._redraw_plot()

    def handle_item_removed(self, index: int, item: ProbePositionsRepositoryItem) -> None:
        self._table_model.remove_item(index, item)



from __future__ import annotations
from collections.abc import Sequence
from enum import auto, IntEnum
import logging

import numpy.random
import scipy.linalg

from ptychodus.api.common import ComplexArrayType, RealArrayType
from ptychodus.api.parametric import ParameterGroup
from ptychodus.api.probe import ProbeSequence, ProbeGeometryProvider
from ptychodus.api.propagator import intensity

from .settings import ProbeSettings

logger = logging.getLogger(__name__)


class ProbeModeDecayType(IntEnum):
    NONE = auto()
    POLYNOMIAL = auto()
    EXPONENTIAL = auto()

    def get_weights(self, num_modes: int, decay_ratio: float) -> Sequence[float]:
        match self.value:
            case ProbeModeDecayType.EXPONENTIAL:
                b = 1.0 + (1.0 - decay_ratio) / decay_ratio
                return [b**-n for n in range(num_modes)]
            case ProbeModeDecayType.POLYNOMIAL:
                b = numpy.log(decay_ratio) / numpy.log(2.0)
                return [(n + 1) ** b for n in range(num_modes)]
            case _:
                return [1.0] + [0.0] * (num_modes - 1)


class MultimodalProbeBuilder(ParameterGroup):
    def __init__(self, rng: numpy.random.Generator, settings: ProbeSettings) -> None:
        super().__init__()
        self._rng = rng
        self._settings = settings

        self.num_incoherent_modes = settings.num_incoherent_modes.copy()
        self._add_parameter('num_incoherent_modes', self.num_incoherent_modes)

        self.orthogonalize_incoherent_modes = settings.orthogonalize_incoherent_modes.copy()
        self._add_parameter('orthogonalize_incoherent_modes', self.orthogonalize_incoherent_modes)

        self.incoherent_mode_decay_type = settings.incoherent_mode_decay_type.copy()
        self._add_parameter('incoherent_mode_decay_type', self.incoherent_mode_decay_type)

        self.incoherent_mode_decay_ratio = settings.incoherent_mode_decay_ratio.copy()
        self._add_parameter('incoherent_mode_decay_ratio', self.incoherent_mode_decay_ratio)

        self.num_coherent_modes = settings.num_coherent_modes.copy()
        self._add_parameter('num_coherent_modes', self.num_coherent_modes)

    def sync_to_settings(self) -> None:
        for parameter in self.parameters().values():
            parameter.sync_value_to_parent()

    def copy(self) -> MultimodalProbeBuilder:
        builder = MultimodalProbeBuilder(self._rng, self._settings)

        for key, value in self.parameters().items():
            builder.parameters()[key].set_value(value.get_value())

        return builder

    def _orthogonalize_incoherent_modes(self, array_in: ComplexArrayType) -> ComplexArrayType:
        array_out = array_in.copy()

        if array_in.shape[-3] > 1:
            imodes_as_rows = array_in[0].reshape(array_in.shape[-3], -1)
            imodes_as_cols = imodes_as_rows.T

            try:
                imodes_as_ortho_cols = scipy.linalg.orth(imodes_as_cols)
            except ValueError as ex:
                logger.exception(ex)
                return array_in.copy()

            imodes_as_ortho_rows = imodes_as_ortho_cols.T
            imodes_ortho = imodes_as_ortho_rows.reshape(*array_in.shape)

            array_out[0, :, :, :] = imodes_ortho

        return array_out

    def _get_imode_weights(self, num_imodes: int) -> Sequence[float]:
        imode_decay_type_text = self.incoherent_mode_decay_type.get_value()
        imode_decay_ratio = self.incoherent_mode_decay_ratio.get_value()
        imode_decay_type = ProbeModeDecayType.NONE

        if imode_decay_ratio > 0.0:
            try:
                imode_decay_type = ProbeModeDecayType[imode_decay_type_text.upper()]
            except KeyError:
                logger.debug(f'Unknown probe mode decay type "{imode_decay_type_text}"')

        return imode_decay_type.get_weights(num_imodes, imode_decay_ratio)

    def _adjust_imode_power(self, array_in: ComplexArrayType, power: float) -> ComplexArrayType:
        imode_weights = self._get_imode_weights(array_in.shape[-3])
        array_out = array_in.copy()
        it = iter(array_out[0])  # iterate incoherent modes

        for weight in imode_weights:
            imode = next(it)
            ipower = numpy.sum(numpy.square(numpy.abs(imode)))
            imode *= numpy.sqrt(weight * power / ipower)

        return array_out

    def _random_phase_shift_axis(self, size: int) -> ComplexArrayType:
        a = self._rng.uniform() - 0.5
        b = (size - 1 - 2 * numpy.arange(size)) / size
        return numpy.exp(1j * numpy.pi * a * b)

    def _init_modes(
        self,
        geometry_provider: ProbeGeometryProvider,
        array_in: ComplexArrayType,
        normalize_cmodes: bool = True,
    ) -> ComplexArrayType:
        assert array_in.ndim == 4
        num_cmodes = self.num_coherent_modes.get_value()
        num_imodes = self.num_incoherent_modes.get_value()
        height = array_in.shape[-2]
        width = array_in.shape[-1]

        array_out = numpy.zeros((num_cmodes, num_imodes, height, width), array_in.dtype)

        for cmode in range(num_cmodes):
            if cmode < array_in.shape[0]:
                # copy existing cmode
                values = array_in[cmode, 0, :, :]
            else:
                # randomize new cmode
                real = self._rng.normal(0.0, 1.0, size=(height, width))
                imag = self._rng.normal(0.0, 1.0, size=(height, width))
                values = real + 1j * imag

                if normalize_cmodes:
                    values /= numpy.sqrt(numpy.mean(intensity(values)))

            array_out[cmode, 0, :, :] = values

        for imode in range(num_imodes):
            if imode < array_in.shape[1]:
                # copy existing imode
                values = array_in[0, imode, :, :]
            else:
                # apply random phase shift to first imode
                first_imode = array_in[0, 0, :, :]
                phase_shift_y = self._random_phase_shift_axis(height)
                phase_shift_x = self._random_phase_shift_axis(width)
                values = first_imode * numpy.outer(phase_shift_y, phase_shift_x)

            array_out[0, imode, :, :] = values

        if self.orthogonalize_incoherent_modes.get_value():
            array_out = self._orthogonalize_incoherent_modes(array_out)

        if geometry_provider.probe_photon_count > 0.0:
            array_out = self._adjust_imode_power(array_out, geometry_provider.probe_photon_count)

        return array_out

    def _init_opr_weights(
        self, geometry_provider: ProbeGeometryProvider, small_value: float = 1.0e-6
    ) -> RealArrayType | None:
        num_scan_points = geometry_provider.num_scan_points
        num_cmodes = self.num_coherent_modes.get_value()
        opr_weights: RealArrayType | None = None

        if self.num_coherent_modes.get_value() > 1:
            opr_weights = small_value * self._rng.normal(
                0.0, 1.0, size=(num_scan_points, num_cmodes)
            )
            assert opr_weights is not None  # unnecessary but makes pylance less annoying
            opr_weights[:, 0] = 1.0

        return opr_weights

    def set_identity(self) -> None:
        self.num_coherent_modes.set_value(1)
        self.num_incoherent_modes.set_value(1)

    def build(
        self, probes: ProbeSequence, geometry_provider: ProbeGeometryProvider
    ) -> ProbeSequence:
        if self.num_coherent_modes.get_value() <= 1 and self.num_incoherent_modes.get_value() <= 1:
            return probes

        array = self._init_modes(geometry_provider, probes.get_array())

        try:
            opr_weights: RealArrayType | None = probes.get_opr_weights()
        except ValueError:
            opr_weights = self._init_opr_weights(geometry_provider)
        else:
            # TODO if opr_weights.shape[0] != num_scan_points: raise ValueError()
            pass

        return ProbeSequence(
            array=array,
            opr_weights=opr_weights,
            pixel_geometry=probes.get_pixel_geometry(),
        )



from __future__ import annotations
from collections.abc import Iterator

import numpy

from ptychodus.api.geometry import AffineTransform
from ptychodus.api.parametric import ParameterGroup
from ptychodus.api.probe_positions import ProbePosition

from .settings import ProbePositionsSettings


class ProbePositionTransform(ParameterGroup):
    def __init__(self, rng: numpy.random.Generator, settings: ProbePositionsSettings) -> None:
        super().__init__()
        self._rng = rng
        self._settings = settings

        self.affine00 = settings.affine00.copy()
        self._add_parameter('affine00', self.affine00)

        self.affine01 = settings.affine01.copy()
        self._add_parameter('affine01', self.affine01)

        self.affine02 = settings.affine02.copy()
        self._add_parameter('affine02', self.affine02)

        self.affine10 = settings.affine10.copy()
        self._add_parameter('affine10', self.affine10)

        self.affine11 = settings.affine11.copy()
        self._add_parameter('affine11', self.affine11)

        self.affine12 = settings.affine12.copy()
        self._add_parameter('affine12', self.affine12)

        self.jitter_radius_m = settings.jitter_radius_m.copy()
        self._add_parameter('jitter_radius_m', self.jitter_radius_m)

    def sync_to_settings(self) -> None:
        for parameter in self.parameters().values():
            parameter.sync_value_to_parent()

    def copy(self) -> ProbePositionTransform:
        transform = ProbePositionTransform(self._rng, self._settings)

        for key, value in self.parameters().items():
            transform.parameters()[key].set_value(value.get_value())

        return transform

    @staticmethod
    def negate_x(preset: int) -> bool:
        return preset & 0x1 != 0x0

    @staticmethod
    def negate_y(preset: int) -> bool:
        return preset & 0x2 != 0x0

    @staticmethod
    def swap_xy(preset: int) -> bool:
        return preset & 0x4 != 0x0

    def labels_for_presets(self) -> Iterator[str]:
        for index in range(8):
            yp = '\u2212y' if self.negate_y(index) else '\u002by'
            xp = '\u2212x' if self.negate_x(index) else '\u002bx'
            fyx = f'{xp}, {yp}' if self.swap_xy(index) else f'{yp}, {xp}'
            yield f'(y, x) \u2192 ({fyx})'

    def apply_presets(self, index: int) -> None:
        self.block_notifications(True)

        if self.swap_xy(index):
            self.affine00.set_value(0)
            self.affine01.set_value(-1 if self.negate_y(index) else +1)
            self.affine02.set_value(0)
            self.affine10.set_value(-1 if self.negate_x(index) else +1)
            self.affine11.set_value(0)
            self.affine12.set_value(0)
        else:
            self.affine00.set_value(-1 if self.negate_x(index) else +1)
            self.affine01.set_value(0)
            self.affine02.set_value(0)
            self.affine10.set_value(0)
            self.affine11.set_value(-1 if self.negate_y(index) else +1)
            self.affine12.set_value(0)

        self.block_notifications(False)

    def get_transform(self) -> AffineTransform:
        return AffineTransform(
            a00=self.affine00.get_value(),
            a01=self.affine01.get_value(),
            a02=self.affine02.get_value(),
            a10=self.affine10.get_value(),
            a11=self.affine11.get_value(),
            a12=self.affine12.get_value(),
        )

    def set_transform(self, transform: AffineTransform) -> None:
        self.block_notifications(True)

        self.affine00.set_value(transform.a00)
        self.affine01.set_value(transform.a01)
        self.affine02.set_value(transform.a02)

        self.affine10.set_value(transform.a10)
        self.affine11.set_value(transform.a11)
        self.affine12.set_value(transform.a12)

        self.block_notifications(False)

    def set_identity(self) -> None:
        self.apply_presets(0)

    def __call__(self, point: ProbePosition) -> ProbePosition:
        transform = self.get_transform()
        pos_x, pos_y = transform(point.coordinate_x_m, point.coordinate_y_m)
        rad = self.jitter_radius_m.get_value()

        if rad > 0.0:
            while True:
                dx = self._rng.uniform()
                dy = self._rng.uniform()

                if dx * dx + dy * dy < 1.0:
                    pos_x += dx * rad
                    pos_y += dy * rad
                    break

        return ProbePosition(point.index, pos_x, pos_y)



"""Product data structure bundling the probe positions, probe sequence, object and metadata."""

from abc import ABC, abstractmethod
from collections.abc import Sequence
from dataclasses import dataclass
from pathlib import Path
from sys import getsizeof

from .common import ELECTRON_VOLT_J, PLANCK_CONSTANT_J_PER_HZ, LIGHT_SPEED_M_PER_S
from .object import Object
from .probe import ProbeSequence
from .probe_positions import ProbePositionSequence


@dataclass(frozen=True)
class ProductMetadata:
    """Metadata for the sample and experiment geometry."""

    name: str
    comments: str
    detector_distance_m: float
    probe_energy_eV: float  # noqa: N815
    probe_photon_count: float
    exposure_time_s: float
    mass_attenuation_m2_kg: float
    tomography_angle_deg: float

    @property
    def probe_energy_J(self) -> float:  # noqa: N802
        return self.probe_energy_eV * ELECTRON_VOLT_J

    @property
    def probe_wavelength_m(self) -> float:
        hc_Jm = PLANCK_CONSTANT_J_PER_HZ * LIGHT_SPEED_M_PER_S  # noqa: N806

        try:
            return hc_Jm / self.probe_energy_J
        except ZeroDivisionError:
            return 0.0

    @property
    def nbytes(self) -> int:
        sz = getsizeof(self.name)
        sz += getsizeof(self.comments)
        sz += getsizeof(self.detector_distance_m)
        sz += getsizeof(self.probe_energy_eV)
        sz += getsizeof(self.probe_photon_count)
        sz += getsizeof(self.exposure_time_s)
        sz += getsizeof(self.mass_attenuation_m2_kg)
        sz += getsizeof(self.tomography_angle_deg)
        return sz


@dataclass(frozen=True)
class LossValue:
    """Loss recorded at a given epoch."""

    epoch: int
    value: float


@dataclass(frozen=True)
class Product:
    """A Data Product bundles metadata, positions, probes, object, and loss history."""

    metadata: ProductMetadata
    probe_positions: ProbePositionSequence
    probes: ProbeSequence
    object_: Object
    losses: Sequence[LossValue]

    @property
    def nbytes(self) -> int:
        sz = self.metadata.nbytes
        sz += self.probe_positions.nbytes
        sz += self.probes.nbytes
        sz += self.object_.nbytes
        return sz


class ProductFileReader(ABC):
    """Plugin interface for reading data products."""

    @abstractmethod
    def read(self, file_path: Path) -> Product:
        """Read a data product from file."""
        pass


class ProductFileWriter(ABC):
    """Plugin interface for writing data products."""

    @abstractmethod
    def write(self, file_path: Path, product: Product) -> None:
        """Write a data product to file."""
        pass



from PyQt5.QtWidgets import (
    QButtonGroup,
    QDialog,
    QFormLayout,
    QGroupBox,
    QHBoxLayout,
    QMessageBox,
    QRadioButton,
    QSpinBox,
    QTableView,
    QWidget,
)

from ptychodus.api.observer import Observable, Observer
from ptychodus.api.parametric import StringParameter

from ...model.product.probe import (
    AveragePatternProbeBuilder,
    DiskProbeBuilder,
    FresnelZonePlateProbeBuilder,
    MultimodalProbeBuilder,
    ProbeModeDecayType,
    ProbeRepositoryItem,
    RectangularProbeBuilder,
    SuperGaussianProbeBuilder,
    ZernikeProbeBuilder,
)
from ...view.widgets import GroupBoxWithPresets
from ..parametric import (
    LengthWidgetParameterViewController,
    ParameterViewBuilder,
    ParameterViewController,
)
from .zernike import ZernikeTableModel

__all__ = [
    'ProbeEditorViewControllerFactory',
]


class FresnelZonePlateViewController(ParameterViewController):
    def __init__(self, title: str, probe_builder: FresnelZonePlateProbeBuilder) -> None:
        super().__init__()
        self._widget = GroupBoxWithPresets(title)

        for label in probe_builder.labels_for_presets():
            action = self._widget.presets_menu.addAction(label)

            if action is None:
                raise ValueError('action is None!')
            else:
                action.triggered.connect(lambda _, label=label: probe_builder.apply_presets(label))

        self._zone_plate_diameter_view_controller = LengthWidgetParameterViewController(
            probe_builder.zone_plate_diameter_m
        )
        self._outermost_zone_width_view_controller = LengthWidgetParameterViewController(
            probe_builder.outermost_zone_width_m
        )
        self._central_beamstop_diameter_view_controller = LengthWidgetParameterViewController(
            probe_builder.central_beamstop_diameter_m
        )
        self._defocus_distance_view_controller = LengthWidgetParameterViewController(
            probe_builder.defocus_distance_m
        )

        layout = QFormLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addRow(
            'Zone Plate Diameter:', self._zone_plate_diameter_view_controller.get_widget()
        )
        layout.addRow(
            'Outermost Zone Width:',
            self._outermost_zone_width_view_controller.get_widget(),
        )
        layout.addRow(
            'Central Beamstop Diameter:',
            self._central_beamstop_diameter_view_controller.get_widget(),
        )
        layout.addRow('Defocus Distance:', self._defocus_distance_view_controller.get_widget())
        self._widget.contents.setLayout(layout)

    def get_widget(self) -> QWidget:
        return self._widget


class ZernikeViewController(ParameterViewController, Observer):
    def __init__(self, title: str, probe_builder: ZernikeProbeBuilder) -> None:
        super().__init__()
        self._widget = QGroupBox(title)
        self._probe_builder = probe_builder
        self._order_spin_box = QSpinBox()
        self._coefficients_table_model = ZernikeTableModel(probe_builder)
        self._coefficients_table_view = QTableView()
        self._diameter_view_controller = LengthWidgetParameterViewController(
            probe_builder.diameter_m
        )

        self._coefficients_table_view.setModel(self._coefficients_table_model)

        layout = QFormLayout()
        layout.addRow('Diameter:', self._diameter_view_controller.get_widget())
        layout.addRow('Order:', self._order_spin_box)
        layout.addRow(self._coefficients_table_view)
        self._widget.setLayout(layout)

        self._sync_model_to_view()
        self._order_spin_box.valueChanged.connect(probe_builder.set_order)
        probe_builder.add_observer(self)

    def get_widget(self) -> QWidget:
        return self._widget

    def _sync_model_to_view(self) -> None:
        self._order_spin_box.setRange(1, 100)
        self._order_spin_box.setValue(self._probe_builder.get_order())

        self._coefficients_table_model.beginResetModel()  # TODO clean up
        self._coefficients_table_model.endResetModel()

    def _update(self, observable: Observable) -> None:
        if observable is self._probe_builder:
            self._sync_model_to_view()


class DecayTypeParameterViewController(ParameterViewController, Observer):
    def __init__(self, parameter: StringParameter) -> None:
        super().__init__()
        self._parameter = parameter
        self._polynomial_decay_button = QRadioButton('Polynomial')
        self._exponential_decay_button = QRadioButton('Exponential')

        self._button_group = QButtonGroup()
        self._button_group.addButton(
            self._polynomial_decay_button, ProbeModeDecayType.POLYNOMIAL.value
        )
        self._button_group.addButton(
            self._exponential_decay_button, ProbeModeDecayType.EXPONENTIAL.value
        )
        self._button_group.setExclusive(True)
        self._button_group.idToggled.connect(self._sync_view_to_model)

        layout = QHBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(self._polynomial_decay_button)
        layout.addWidget(self._exponential_decay_button)

        self._widget = QWidget()
        self._widget.setLayout(layout)

        self._sync_model_to_view()
        parameter.add_observer(self)

    def get_widget(self) -> QWidget:
        return self._widget

    def _sync_view_to_model(self, tool_id: int, checked: bool) -> None:
        if checked:
            decay_type = ProbeModeDecayType(tool_id)
            self._parameter.set_value(decay_type.name)

    def _sync_model_to_view(self) -> None:
        try:
            decay_type = ProbeModeDecayType[self._parameter.get_value().upper()]
        except KeyError:
            decay_type = ProbeModeDecayType.POLYNOMIAL

        button = self._button_group.button(decay_type.value)

        if button is None:
            raise ValueError('button is None!')
        else:
            button.setChecked(True)

    def _update(self, observable: Observable) -> None:
        if observable is self._parameter:
            self._sync_model_to_view()


class ProbeEditorViewControllerFactory:
    def _append_additional_modes(
        self,
        dialog_builder: ParameterViewBuilder,
        additional_modes_builder: MultimodalProbeBuilder | None,
    ) -> None:
        if additional_modes_builder is None:
            return

        incoherent_modes_group = 'Incoherent (Mixed State) Modes'
        dialog_builder.add_spin_box(
            additional_modes_builder.num_incoherent_modes,
            'Number of Modes:',
            group=incoherent_modes_group,
        )
        dialog_builder.add_check_box(
            additional_modes_builder.orthogonalize_incoherent_modes,
            'Orthogonalize Modes:',
            group=incoherent_modes_group,
        )
        dialog_builder.add_view_controller(
            DecayTypeParameterViewController(additional_modes_builder.incoherent_mode_decay_type),
            'Decay Type:',
            group=incoherent_modes_group,
        )
        dialog_builder.add_decimal_slider(
            additional_modes_builder.incoherent_mode_decay_ratio,
            'Decay Ratio:',
            group=incoherent_modes_group,
        )

        coherent_modes_group = 'Coherent (OPR) Modes'
        dialog_builder.add_spin_box(
            additional_modes_builder.num_coherent_modes,
            'Number of Modes:',
            group=coherent_modes_group,
        )

    def create_editor_dialog(
        self, item_name: str, item: ProbeRepositoryItem, parent: QWidget
    ) -> QDialog:
        probe_builder = item.get_builder()
        builder_name = probe_builder.get_name()
        additional_modes_builder = item.get_additional_modes_builder()
        primary_mode_group = 'Primary Mode'
        title = f'{item_name} [{builder_name}]'

        if isinstance(probe_builder, AveragePatternProbeBuilder):
            dialog_builder = ParameterViewBuilder()
            self._append_additional_modes(dialog_builder, additional_modes_builder)
            return dialog_builder.build_dialog(title, parent)
        elif isinstance(probe_builder, DiskProbeBuilder):
            dialog_builder = ParameterViewBuilder()
            dialog_builder.add_length_widget(
                probe_builder.diameter_m,
                'Diameter:',
                group=primary_mode_group,
            )
            dialog_builder.add_length_widget(
                probe_builder.defocus_distance_m,
                'Defocus Distance:',
                group=primary_mode_group,
            )
            self._append_additional_modes(dialog_builder, additional_modes_builder)
            return dialog_builder.build_dialog(title, parent)
        elif isinstance(probe_builder, FresnelZonePlateProbeBuilder):
            dialog_builder = ParameterViewBuilder()
            dialog_builder.add_view_controller_to_top(
                FresnelZonePlateViewController(primary_mode_group, probe_builder)
            )
            self._append_additional_modes(dialog_builder, additional_modes_builder)
            return dialog_builder.build_dialog(title, parent)
        elif isinstance(probe_builder, RectangularProbeBuilder):
            dialog_builder = ParameterViewBuilder()
            dialog_builder.add_length_widget(
                probe_builder.width_m,
                'Width:',
                group=primary_mode_group,
            )
            dialog_builder.add_length_widget(
                probe_builder.height_m,
                'Height:',
                group=primary_mode_group,
            )
            dialog_builder.add_length_widget(
                probe_builder.defocus_distance_m,
                'Defocus Distance:',
                group=primary_mode_group,
            )
            self._append_additional_modes(dialog_builder, additional_modes_builder)
            return dialog_builder.build_dialog(title, parent)
        elif isinstance(probe_builder, SuperGaussianProbeBuilder):
            dialog_builder = ParameterViewBuilder()
            dialog_builder.add_length_widget(
                probe_builder.annular_radius_m,
                'Annular Radius:',
                group=primary_mode_group,
            )
            dialog_builder.add_length_widget(
                probe_builder.fwhm_m,
                'Full Width at Half Maximum:',
                group=primary_mode_group,
            )
            dialog_builder.add_decimal_line_edit(
                probe_builder.order_parameter,
                'Order Parameter:',
                group=primary_mode_group,
            )
            self._append_additional_modes(dialog_builder, additional_modes_builder)
            return dialog_builder.build_dialog(title, parent)
        elif isinstance(probe_builder, ZernikeProbeBuilder):
            dialog_builder = ParameterViewBuilder()
            dialog_builder.add_view_controller_to_top(
                ZernikeViewController(primary_mode_group, probe_builder)
            )
            self._append_additional_modes(dialog_builder, additional_modes_builder)
            return dialog_builder.build_dialog(title, parent)

        return QMessageBox(
            QMessageBox.Icon.Information,
            title,
            f'"{builder_name}" has no editable parameters!',
            QMessageBox.StandardButton.Ok,
            parent,
        )



import logging

import numpy

from ptychodus.api.common import RealArrayType
from ptychodus.api.geometry import Interval
from ptychodus.api.parametric import ParameterGroup

logger = logging.getLogger(__name__)


class ColorAxis(ParameterGroup):
    def __init__(self) -> None:
        super().__init__()
        self.lower = self.create_real_parameter('lower', 0.0)
        self.upper = self.create_real_parameter('upper', 1.0)

    def get_range(self) -> Interval[float]:
        return Interval[float].create_proper(
            self.lower.get_value(),
            self.upper.get_value(),
        )

    def set_range(self, lower: float, upper: float):
        self.lower.set_value(lower, notify=False)
        self.upper.set_value(upper, notify=False)
        self.notify_observers()

    def set_to_data_range(self, array: RealArrayType) -> None:
        if array.size > 0:
            lower = array.min().item()
            upper = array.max().item()

            if numpy.isfinite(lower) and numpy.isfinite(upper):
                if lower == upper:
                    logger.debug('Array values are uniform.')
                    lower -= 0.5
                    upper += 0.5
            else:
                logger.warning('Array values not finite!')
                lower = 0.0
                upper = 1.0

            self.set_range(lower, upper)
        else:
            logger.warning('Array has zero size!')



from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import (
    QComboBox,
    QDialog,
    QFormLayout,
    QGridLayout,
    QGroupBox,
    QHBoxLayout,
    QLabel,
    QListView,
    QPushButton,
    QRadioButton,
    QSlider,
    QSpinBox,
    QStackedWidget,
    QStatusBar,
    QVBoxLayout,
    QWidget,
)

from .visualization import VisualizationParametersView, VisualizationWidget
from .widgets import DecimalLineEdit, LengthWidget


class ProbePropagationParametersView(QGroupBox):
    def __init__(self, parent: QWidget | None = None) -> None:
        super().__init__('Parameters', parent)
        self.begin_coordinate_widget = LengthWidget(is_signed=True)
        self.end_coordinate_widget = LengthWidget(is_signed=True)
        self.num_steps_spin_box = QSpinBox()
        self.visualization_parameters_view = VisualizationParametersView()

        propagation_layout = QFormLayout()
        propagation_layout.addRow('Begin Coordinate:', self.begin_coordinate_widget)
        propagation_layout.addRow('End Coordinate:', self.end_coordinate_widget)
        propagation_layout.addRow('Number of Steps:', self.num_steps_spin_box)

        propagation_group_box = QGroupBox('Propagation')
        propagation_group_box.setLayout(propagation_layout)

        layout = QVBoxLayout()
        layout.addWidget(propagation_group_box)
        layout.addWidget(self.visualization_parameters_view)
        layout.addStretch()
        self.setLayout(layout)


class ProbePropagationDialog(QDialog):
    def __init__(self, parent: QWidget | None = None) -> None:
        super().__init__(parent)
        self.xy_view = VisualizationWidget('XY Plane')
        self.zx_view = VisualizationWidget('ZX Plane')
        self.parameters_view = ProbePropagationParametersView()
        self.zy_view = VisualizationWidget('ZY Plane')
        self.propagate_button = QPushButton('Propagate')
        self.save_button = QPushButton('Save')
        self.coordinate_slider = QSlider(Qt.Orientation.Horizontal)
        self.coordinate_label = QLabel()
        self.status_bar = QStatusBar()

        action_layout = QHBoxLayout()
        action_layout.addWidget(self.propagate_button)
        action_layout.addWidget(self.save_button)

        coordinate_layout = QHBoxLayout()
        coordinate_layout.setContentsMargins(0, 0, 0, 0)
        coordinate_layout.addWidget(self.coordinate_slider)
        coordinate_layout.addWidget(self.coordinate_label)

        contents_layout = QGridLayout()
        contents_layout.addWidget(self.xy_view, 0, 0)
        contents_layout.addWidget(self.zx_view, 0, 1)
        contents_layout.addWidget(self.parameters_view, 1, 0)
        contents_layout.addWidget(self.zy_view, 1, 1)
        contents_layout.addLayout(action_layout, 2, 0)
        contents_layout.addLayout(coordinate_layout, 2, 1)
        contents_layout.setColumnStretch(0, 1)
        contents_layout.setColumnStretch(1, 2)

        layout = QVBoxLayout()
        layout.addLayout(contents_layout)
        layout.addWidget(self.status_bar)
        self.setLayout(layout)


class IlluminationParametersView(QGroupBox):
    def __init__(self, parent: QWidget | None = None) -> None:
        super().__init__('Parameters', parent)
        self.photon_flux_line_edit = DecimalLineEdit.create_instance()
        self.exposure_time_line_edit = DecimalLineEdit.create_instance()
        self.mass_attenuation_label = QLabel('Mass Attenuation [m\u00b2/kg]:')
        self.mass_attenuation_line_edit = DecimalLineEdit.create_instance()

        layout = QFormLayout()
        layout.addRow('Photon Flux [ph/s]:', self.photon_flux_line_edit)
        layout.addRow('Exposure Time [s]:', self.exposure_time_line_edit)
        layout.addRow(self.mass_attenuation_label)
        layout.addRow(self.mass_attenuation_line_edit)
        self.setLayout(layout)


class IlluminationQuantityView(QGroupBox):
    def __init__(self, parent: QWidget | None = None) -> None:
        super().__init__('Quantity', parent)
        self.photon_number_button = QRadioButton('Photon Number')
        self.photon_fluence_button = QRadioButton('Photon Fluence [1/m\u00b2]')
        self.photon_fluence_rate_button = QRadioButton('Photon Fluence Rate [Hz/m\u00b2]')
        self.energy_fluence_button = QRadioButton('Energy Fluence [J/m\u00b2]')
        self.energy_fluence_rate_button = QRadioButton('Energy Fluence Rate [W/m\u00b2]')
        self.dose_button = QRadioButton('Dose [Gy]')
        self.dose_rate_button = QRadioButton('Dose Rate [Gy/s]')

        layout = QVBoxLayout()
        layout.addWidget(self.photon_number_button)
        layout.addWidget(self.photon_fluence_button)
        layout.addWidget(self.photon_fluence_rate_button)
        layout.addWidget(self.energy_fluence_button)
        layout.addWidget(self.energy_fluence_rate_button)
        layout.addWidget(self.dose_button)
        layout.addWidget(self.dose_rate_button)
        self.setLayout(layout)


class IlluminationDialog(QDialog):
    def __init__(self, parent: QWidget | None = None) -> None:
        super().__init__(parent)
        self.visualization_widget = VisualizationWidget('Visualization')
        self.parameters_view = IlluminationParametersView()
        self.quantity_view = IlluminationQuantityView()
        self.visualization_parameters_view = VisualizationParametersView()
        self.save_button = QPushButton('Save')
        self.status_bar = QStatusBar()

        parameter_layout = QVBoxLayout()
        parameter_layout.addWidget(self.parameters_view)
        parameter_layout.addWidget(self.quantity_view)
        parameter_layout.addWidget(self.visualization_parameters_view)
        parameter_layout.addWidget(self.save_button)
        parameter_layout.addStretch()

        contents_layout = QHBoxLayout()
        contents_layout.addWidget(self.visualization_widget, 1)
        contents_layout.addLayout(parameter_layout)

        layout = QVBoxLayout()
        layout.addLayout(contents_layout)
        layout.addWidget(self.status_bar)
        self.setLayout(layout)


class FluorescenceVSPIParametersView(QWidget):
    def __init__(self, parent: QWidget | None = None) -> None:
        super().__init__(parent)
        self.damping_factor_line_edit = DecimalLineEdit.create_instance()
        self.max_iterations_spin_box = QSpinBox()

        layout = QFormLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addRow('Damping Factor:', self.damping_factor_line_edit)
        layout.addRow('Max Iterations:', self.max_iterations_spin_box)
        self.setLayout(layout)


class FluorescenceTwoStepParametersView(QWidget):
    def __init__(self, parent: QWidget | None = None) -> None:
        super().__init__(parent)
        self.upscaling_strategy_combo_box = QComboBox()
        self.deconvolution_strategy_combo_box = QComboBox()

        layout = QFormLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addRow('Upscaling Strategy:', self.upscaling_strategy_combo_box)
        layout.addRow('Deconvolution Strategy:', self.deconvolution_strategy_combo_box)
        self.setLayout(layout)


class FluorescenceParametersView(QGroupBox):
    def __init__(self, parent: QWidget | None = None) -> None:
        super().__init__('Enhancement Strategy', parent)
        self.open_button = QPushButton('Open Measured Dataset')
        self.algorithm_combo_box = QComboBox()
        self.stacked_widget = QStackedWidget()
        self.enhance_button = QPushButton('Enhance')
        self.save_button = QPushButton('Save Enhanced Dataset')

        stacked_widget_layout = self.stacked_widget.layout()

        if stacked_widget_layout is not None:
            stacked_widget_layout.setContentsMargins(0, 0, 0, 0)

        layout = QFormLayout()
        layout.addRow(self.open_button)
        layout.addRow('Algorithm:', self.algorithm_combo_box)
        layout.addRow(self.stacked_widget)
        layout.addRow(self.enhance_button)
        layout.addRow(self.save_button)
        self.setLayout(layout)


class FluorescenceDialog(QDialog):
    def __init__(self, parent: QWidget | None = None) -> None:
        super().__init__(parent)
        self.measured_widget = VisualizationWidget('Measured')
        self.enhanced_widget = VisualizationWidget('Enhanced')
        self.fluorescence_parameters_view = FluorescenceParametersView()
        self.fluorescence_channel_list_view = QListView()
        self.visualization_parameters_view = VisualizationParametersView()
        self.status_bar = QStatusBar()

        parameter_layout = QVBoxLayout()
        parameter_layout.addWidget(self.fluorescence_parameters_view)
        parameter_layout.addWidget(self.fluorescence_channel_list_view, 1)
        parameter_layout.addWidget(self.visualization_parameters_view)
        parameter_layout.addStretch()

        contents_layout = QHBoxLayout()
        contents_layout.addWidget(self.measured_widget, 1)
        contents_layout.addWidget(self.enhanced_widget, 1)
        contents_layout.addLayout(parameter_layout)

        layout = QVBoxLayout()
        layout.addLayout(contents_layout)
        layout.addWidget(self.status_bar)
        self.setLayout(layout)



from collections.abc import Iterator
from importlib.metadata import PackageNotFoundError, version
import logging

from ptychodus.api.reconstructor import (
    NullReconstructor,
    Reconstructor,
    ReconstructorLibrary,
    TrainableReconstructor,
)
from ptychodus.api.settings import SettingsRegistry

from .enums import PtychoPINNEnumerators
from .settings import (
    PtychoPINNInferenceSettings,
    PtychoPINNModelSettings,
    PtychoPINNTrainingSettings,
)

logger = logging.getLogger(__name__)


class PtychoPINNReconstructorLibrary(ReconstructorLibrary):
    def __init__(
        self, settings_registry: SettingsRegistry, is_developer_mode_enabled: bool
    ) -> None:
        super().__init__('ptychopinn')
        self.model_settings = PtychoPINNModelSettings(settings_registry)
        self.training_settings = PtychoPINNTrainingSettings(settings_registry)
        self.inference_settings = PtychoPINNInferenceSettings(settings_registry)
        self.enumerators = PtychoPINNEnumerators()
        self._reconstructors: list[TrainableReconstructor] = list()

        try:
            from .reconstructor import PtychoPINNTrainableReconstructor
        except ModuleNotFoundError:
            logger.info('PtychoPINN not found.')

            if is_developer_mode_enabled:
                for reconstructor in ('PINN', 'Supervised'):
                    self._reconstructors.append(NullReconstructor(reconstructor))
        else:
            try:
                ptychopinn_version = version('ptychopinn')
            except PackageNotFoundError:
                ptychopinn_version = version('ptycho')

            logger.info(f'\tPtychoPINN {ptychopinn_version}')

            self._reconstructors.append(
                PtychoPINNTrainableReconstructor(
                    'PINN',
                    self.model_settings,
                    self.inference_settings,
                    self.training_settings,
                    is_developer_mode_enabled=is_developer_mode_enabled,
                )
            )
            self._reconstructors.append(
                PtychoPINNTrainableReconstructor(
                    'Supervised',
                    self.model_settings,
                    self.inference_settings,
                    self.training_settings,
                    is_developer_mode_enabled=is_developer_mode_enabled,
                )
            )

    @property
    def name(self) -> str:
        return 'PtychoPINN'

    def __iter__(self) -> Iterator[Reconstructor]:
        return iter(self._reconstructors)



from __future__ import annotations
from collections.abc import Sequence
from importlib.metadata import version
from pathlib import Path
from types import TracebackType
from typing import overload
import logging
import sys

try:
    # NOTE must import hdf5plugin before h5py
    import hdf5plugin  # noqa
except ModuleNotFoundError:
    pass

import h5py
import numpy

from ptychodus.api.diffraction import DiffractionMetadata, DiffractionArray
from ptychodus.api.io import StandardFileLayout
from ptychodus.api.plugins import PluginRegistry
from ptychodus.api.settings import SettingsRegistry
from ptychodus.api.workflow import WorkflowAPI

from .agent import AgentCore
from .analysis import AnalysisCore
from .automation import AutomationCore
from .diffraction import DiffractionCore, PatternsStreamingContext
from .fluorescence import FluorescenceCore
from .globus import GlobusCore
from .memory import MemoryPresenter
from .metadata import MetadataPresenter
from .product import PositionsStreamingContext, ProductCore
from .ptychi import PtyChiReconstructorLibrary
from .ptychonn import PtychoNNReconstructorLibrary
from .ptychopinn import PtychoPINNReconstructorLibrary
from .processing import ProcessingCore
from .task_manager import TaskManager
from .visualization import VisualizationEngine
from .workflow import ConcreteWorkflowAPI

logger = logging.getLogger(__name__)


def configure_logger(*, log_level: int) -> None:
    logging.basicConfig(
        format='%(asctime)s [%(levelname)s] %(name)s: %(message)s',
        stream=sys.stdout,
        encoding='utf-8',
        level=log_level,
    )
    logging.getLogger('matplotlib').setLevel(logging.WARNING)

    logger.info(f'Ptychodus {version("ptychodus")}')
    logger.info(f'NumPy {version("numpy")}')
    logger.info(f'Matplotlib {version("matplotlib")}')
    logger.info(f'HDF5Plugin {version("hdf5plugin")}')
    logger.info(f'H5Py {version("h5py")}')
    logger.info(f'HDF5 {h5py.version.hdf5_version}')


class PtychodusStreamingContext:
    def __init__(
        self,
        positions_context: PositionsStreamingContext,
        patterns_context: PatternsStreamingContext,
    ) -> None:
        self._positions_context = positions_context
        self._patterns_context = patterns_context

    def start(self) -> None:
        self._positions_context.start()
        self._patterns_context.start()

    def append_positions_x(self, values_m: Sequence[float], trigger_counts: Sequence[int]) -> None:
        self._positions_context.append_positions_x(values_m, trigger_counts)

    def append_positions_y(self, values_m: Sequence[float], trigger_counts: Sequence[int]) -> None:
        self._positions_context.append_positions_y(values_m, trigger_counts)

    def append_array(self, array: DiffractionArray) -> None:
        self._patterns_context.append_array(array)

    def get_queue_size(self) -> int:
        return 0  # TODO get_queue_size

    def stop(self) -> None:
        self._patterns_context.stop()
        self._positions_context.stop()


class ModelCore:
    def __init__(
        self, settings_file: Path | None = None, *, log_level: int = logging.WARNING
    ) -> None:
        configure_logger(log_level=log_level)

        self.rng = numpy.random.default_rng()
        self.plugin_registry = PluginRegistry.load_plugins()
        self._task_manager = TaskManager()

        self.memory_presenter = MemoryPresenter()
        self.settings_registry = SettingsRegistry()

        self.diffraction_core = DiffractionCore(
            self._task_manager,
            self.settings_registry,
            self.plugin_registry.bad_pixels_file_readers,
            self.plugin_registry.diffraction_file_readers,
            self.plugin_registry.diffraction_file_writers,
            self.settings_registry,
        )
        self.product_core = ProductCore(
            self.rng,
            self.settings_registry,
            self.diffraction_core.pattern_sizer,
            self.diffraction_core.diffraction_api,
            self.diffraction_core.dataset,
            self.plugin_registry.probe_position_file_readers,
            self.plugin_registry.probe_position_file_writers,
            self.plugin_registry.fresnel_zone_plates,
            self.plugin_registry.probe_file_readers,
            self.plugin_registry.probe_file_writers,
            self.plugin_registry.object_file_readers,
            self.plugin_registry.object_file_writers,
            self.plugin_registry.product_file_readers,
            self.plugin_registry.product_file_writers,
            self.settings_registry,
        )
        self.metadata_presenter = MetadataPresenter(
            self.diffraction_core.detector_settings,
            self.diffraction_core.diffraction_settings,
            self.diffraction_core.dataset,
            self.product_core.settings,
        )

        self.pattern_visualization_engine = VisualizationEngine(is_complex=False)
        self.probe_visualization_engine = VisualizationEngine(is_complex=True)
        self.object_visualization_engine = VisualizationEngine(is_complex=True)

        self.ptychi_reconstructor_library = PtyChiReconstructorLibrary(
            self.settings_registry,
            self.diffraction_core.pattern_sizer,
            self.is_developer_mode_enabled,
        )
        self.ptychonn_reconstructor_library = PtychoNNReconstructorLibrary(
            self.settings_registry, self.is_developer_mode_enabled
        )
        self.ptychopinn_reconstructor_library = PtychoPINNReconstructorLibrary(
            self.settings_registry, self.is_developer_mode_enabled
        )
        self.processing_core = ProcessingCore(
            self._task_manager,
            self.settings_registry,
            self.diffraction_core.diffraction_api,
            self.product_core.product_api,
            [
                self.ptychi_reconstructor_library,
                self.ptychonn_reconstructor_library,
                self.ptychopinn_reconstructor_library,
            ],
        )
        self.fluorescence_core = FluorescenceCore(
            self.settings_registry,
            self.product_core.product_repository,
            self.plugin_registry.upscaling_strategies,
            self.plugin_registry.deconvolution_strategies,
            self.plugin_registry.fluorescence_file_readers,
            self.plugin_registry.fluorescence_file_writers,
        )
        self.analysis_core = AnalysisCore(
            self.settings_registry,
            self.diffraction_core.dataset,
            self.product_core.product_repository,
            self.product_core.object_repository,
        )
        self.globus_core = GlobusCore(
            self._task_manager,
            self.settings_registry,
            self.diffraction_core.diffraction_api,
            self.product_core.product_api,
            self.processing_core.processing_api,
        )
        self.workflow_api: WorkflowAPI = ConcreteWorkflowAPI(
            self.settings_registry,
            self.diffraction_core.diffraction_api,
            self.product_core.product_api,
            self.product_core.probe_positions_api,
            self.product_core.probe_api,
            self.product_core.object_api,
            self.processing_core.processing_api,
            self.globus_core.executor,
        )
        self.automation_core = AutomationCore(
            self._task_manager,
            self.settings_registry,
            self.workflow_api,
            self.plugin_registry.file_based_workflows,
        )
        self.agent_core = AgentCore(self.settings_registry)

        if settings_file:
            self.settings_registry.open_settings(settings_file)

    def __enter__(self) -> ModelCore:
        self._task_manager.start()
        self.globus_core.start()
        self.automation_core.start()
        return self

    @overload
    def __exit__(self, exception_type: None, exception_value: None, traceback: None) -> None: ...

    @overload
    def __exit__(
        self,
        exception_type: type[BaseException],
        exception_value: BaseException,
        traceback: TracebackType,
    ) -> None: ...

    def __exit__(
        self,
        exception_type: type[BaseException] | None,
        exception_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> None:
        self.automation_core.stop()
        self.globus_core.stop()
        self._task_manager.stop(await_finish=False)

    def create_streaming_context(self, metadata: DiffractionMetadata) -> PtychodusStreamingContext:
        return PtychodusStreamingContext(
            self.product_core.probe_positions_api.create_streaming_context(),
            self.diffraction_core.diffraction_api.create_streaming_context(metadata),
        )

    def run_tasks(self) -> None:
        self._task_manager.run_foreground_tasks()

    def _batch_mode_train(self, input_directory: Path, output_directory: Path) -> int:
        product_in_path = input_directory / StandardFileLayout.PRODUCT_IN

        if product_in_path.is_file():
            product_out_path = output_directory / StandardFileLayout.PRODUCT_OUT

            if product_out_path.is_file():
                logger.warning('Output product file will be overwritten!')

            input_product_api = self.workflow_api.load_product(product_in_path)
            input_product_api.train_reconstructor_local(input_directory, output_directory)
            return 0
        else:
            logger.error('Input product is not a file!')
            return -1

    def _batch_mode_reconstruct(self, input_directory: Path, output_directory: Path) -> int:
        settings_path = input_directory / StandardFileLayout.SETTINGS

        if settings_path.is_file():
            self.settings_registry.open_settings(settings_path)
        else:
            logger.warning('Settings file not found! Proceeding with defaults.')

        diffraction_path = input_directory / StandardFileLayout.DIFFRACTION

        if diffraction_path.is_file():
            self.workflow_api.load_assembled_diffraction_data(diffraction_path)
        else:
            logger.error('Diffraction data is not a file!')
            return -1

        product_in_path = input_directory / StandardFileLayout.PRODUCT_IN

        if product_in_path.is_file():
            product_out_path = output_directory / StandardFileLayout.PRODUCT_OUT

            if product_out_path.is_file():
                logger.warning('Output product file will be overwritten!')

            input_product_api = self.workflow_api.load_product(product_in_path)
            output_product_api = input_product_api.reconstruct_local(
                output_product_file=product_out_path, block=True
            )
        else:
            logger.error('Input product is not a file!')
            return -1

        fluorescence_in_path = input_directory / StandardFileLayout.FLUORESCENCE_IN

        if fluorescence_in_path.is_file():
            fluorescence_out_path = output_directory / StandardFileLayout.FLUORESCENCE_OUT

            if fluorescence_out_path.is_file():
                logger.warning('Output fluorescence file will be overwritten!')

            logger.info('Enhancing fluorescence...')

            # TODO add to workflow API
            self.fluorescence_core.enhance_fluorescence(
                output_product_api.get_product_index(),
                fluorescence_in_path,
                fluorescence_out_path,
            )
        else:
            logger.info('No fluorescence data to enhance.')

        return 0

    def batch_mode_execute(self, action: str, input_directory: Path, output_directory: Path) -> int:
        if not input_directory.is_dir():
            raise ValueError('Input path is not a directory!')

        output_directory.mkdir(parents=True, exist_ok=True)

        match action.lower():
            case 'reconstruct':
                return self._batch_mode_reconstruct(input_directory, output_directory)
            case 'train':
                return self._batch_mode_train(input_directory, output_directory)
            case 'infer':
                return self._batch_mode_reconstruct(input_directory, output_directory)

        logger.error(f'Unknown batch mode action "{action}"!')
        return -1

    @property
    def is_developer_mode_enabled(self) -> bool:
        return logger.getEffectiveLevel() <= logging.DEBUG



from pathlib import Path
from typing import Final, Sequence
import logging

import numpy

from ptychodus.api.common import ELECTRON_VOLT_J, LIGHT_SPEED_M_PER_S, PLANCK_CONSTANT_J_PER_HZ
from ptychodus.api.geometry import ImageExtent
from ptychodus.api.object import Object
from ptychodus.api.diffraction import (
    DiffractionDataset,
    DiffractionFileReader,
    DiffractionMetadata,
    SimpleDiffractionDataset,
    SimpleDiffractionArray,
)
from ptychodus.api.plugins import PluginRegistry
from ptychodus.api.probe import ProbeSequence
from ptychodus.api.product import LossValue, Product, ProductFileReader, ProductMetadata
from ptychodus.api.probe_positions import ProbePositionSequence, ProbePosition
from ptychodus.api.tree import SimpleTreeNode

logger = logging.getLogger(__name__)

DEFAULT_DETECTOR_DISTANCE_M = 0.75
DEFAULT_PROBE_ENERGY_EV = 8000.0
DEFAULT_DETECTOR_PIXEL_SIZE_M = 75e-6
PIXEL_COORDINATE_THRESHOLD_M = 1e-3


def _get_scalar(npz_file: numpy.lib.npyio.NpzFile, key: str, fallback: float) -> float:
    if key not in npz_file.files:
        return fallback

    try:
        return float(npz_file[key])
    except (TypeError, ValueError):
        logger.warning('Failed to parse %s from NPZ; using default %s', key, fallback)
        return fallback


def _probe_wavelength_m(energy_ev: float) -> float:
    if energy_ev <= 0.0:
        return 0.0

    hc_jm = PLANCK_CONSTANT_J_PER_HZ * LIGHT_SPEED_M_PER_S
    return hc_jm / (energy_ev * ELECTRON_VOLT_J)


def _coords_in_pixels(scan_x: numpy.ndarray, scan_y: numpy.ndarray) -> bool:
    max_abs = max(float(numpy.max(numpy.abs(scan_x))), float(numpy.max(numpy.abs(scan_y))))
    return max_abs > PIXEL_COORDINATE_THRESHOLD_M


class SLACDiffractionFileReader(DiffractionFileReader):
    def read(self, file_path: Path) -> DiffractionDataset:
        with numpy.load(file_path) as npz_file:
            patterns = numpy.transpose(npz_file['diffraction'], [2, 0, 1])

        num_patterns, detector_height, detector_width = patterns.shape

        metadata = DiffractionMetadata(
            num_patterns_per_array=[num_patterns],
            pattern_dtype=patterns.dtype,
            detector_extent=ImageExtent(detector_width, detector_height),
            file_path=file_path,
        )

        contents_tree = SimpleTreeNode.create_root(['Name', 'Type', 'Details'])
        contents_tree.create_child(
            [file_path.stem, type(patterns).__name__, f'{patterns.dtype}{patterns.shape}']
        )

        array = SimpleDiffractionArray(
            label=file_path.stem,
            indexes=numpy.arange(num_patterns),
            patterns=patterns,
        )

        return SimpleDiffractionDataset(metadata, contents_tree, [array])


class SLACProductFileReader(ProductFileReader):
    def read(self, file_path: Path) -> Product:
        with numpy.load(file_path) as npz_file:
            scan_x = numpy.asarray(npz_file['xcoords_start'])
            scan_y = numpy.asarray(npz_file['ycoords_start'])
            probe_array = npz_file['probeGuess']
            object_array = npz_file['objectGuess']

            detector_distance_m = _get_scalar(
                npz_file, 'detector_distance_m', DEFAULT_DETECTOR_DISTANCE_M
            )
            probe_energy_ev = _get_scalar(npz_file, 'probe_energy_eV', DEFAULT_PROBE_ENERGY_EV)
            if probe_energy_ev == DEFAULT_PROBE_ENERGY_EV:
                probe_energy_ev = _get_scalar(npz_file, 'probe_energy_ev', DEFAULT_PROBE_ENERGY_EV)
            detector_pixel_size_m = _get_scalar(
                npz_file, 'detector_pixel_size_m', DEFAULT_DETECTOR_PIXEL_SIZE_M
            )

        scan_x_m = scan_x
        scan_y_m = scan_y

        if _coords_in_pixels(scan_x, scan_y):
            width_px = float(probe_array.shape[-1])
            detector_width_m = detector_pixel_size_m * width_px
            pixel_size_m = 0.0
            if detector_width_m > 0.0:
                pixel_size_m = _probe_wavelength_m(probe_energy_ev) * detector_distance_m
                pixel_size_m /= detector_width_m

            if pixel_size_m > 0.0:
                scan_x_m = scan_x * pixel_size_m
                scan_y_m = scan_y * pixel_size_m
            else:
                logger.warning('SLAC NPZ coordinate scaling skipped (invalid pixel size).')

        metadata = ProductMetadata(
            name=file_path.stem,
            comments='',
            detector_distance_m=detector_distance_m,
            probe_energy_eV=probe_energy_ev,
            probe_photon_count=0.0,  # not included in file
            exposure_time_s=0.0,  # not included in file
            mass_attenuation_m2_kg=0.0,  # not included in file
            tomography_angle_deg=0.0,  # not included in file
        )

        point_list: list[ProbePosition] = list()

        for idx, (x_m, y_m) in enumerate(zip(scan_x_m, scan_y_m)):
            point = ProbePosition(idx, x_m, y_m)
            point_list.append(point)

        loss: Sequence[LossValue] = list()  # not included in file

        return Product(
            metadata=metadata,
            probe_positions=ProbePositionSequence(point_list),
            probes=ProbeSequence(array=probe_array, opr_weights=None, pixel_geometry=None),
            object_=Object(array=object_array, pixel_geometry=None, center=None),
            losses=loss,
        )


def register_plugins(registry: PluginRegistry) -> None:
    SIMPLE_NAME: Final[str] = 'SLAC_NPZ'  # noqa: N806
    DISPLAY_NAME: Final[str] = 'SLAC NumPy Zipped Archive (*.npz)'  # noqa: N806

    registry.diffraction_file_readers.register_plugin(
        SLACDiffractionFileReader(),
        simple_name=SIMPLE_NAME,
        display_name=DISPLAY_NAME,
    )
    registry.register_product_file_reader_with_adapters(
        SLACProductFileReader(),
        simple_name=SIMPLE_NAME,
        display_name=DISPLAY_NAME,
    )



from pathlib import Path
import logging

import h5py
import numpy

from ptychodus.api.geometry import ImageExtent, PixelGeometry
from ptychodus.api.diffraction import (
    CropCenter,
    DiffractionDataset,
    DiffractionFileReader,
    DiffractionMetadata,
    SimpleDiffractionDataset,
)
from ptychodus.api.plugins import PluginRegistry

from .h5_diffraction_file import H5DiffractionPatternArray, H5DiffractionFileTreeBuilder

logger = logging.getLogger(__name__)


class LYNXDiffractionFileReader(DiffractionFileReader):
    def __init__(self) -> None:
        self._data_path = '/entry/data/eiger_4'
        self._tree_builder = H5DiffractionFileTreeBuilder()

    def read(self, file_path: Path) -> DiffractionDataset:
        with h5py.File(file_path, 'r') as h5_file:
            data = h5_file[self._data_path]

            if not isinstance(data, h5py.Dataset):
                raise KeyError(f'{self._data_path} is not a Dataset!')

            num_patterns, detector_height, detector_width = data.shape

            crop_center: CropCenter | None = None
            detector_distance_m: float | None = None
            detector_pixel_geometry: PixelGeometry | None = None
            exposure_time_s: float | None = None
            probe_energy_eV: float | None = None  # noqa: N806

            try:
                center_x_px: int = data.attrs['Center_x_pixel'].item()
                center_y_px: int = data.attrs['Center_y_pixel'].item()
                detector_distance_m = data.attrs['Detector_distance_m'].item()
                exposure_time_s = data.attrs['Exposure_time'].item()
                photon_energy_keV = data.attrs['Photon_energy_kev'].item()  # noqa: N806
                pixel_size = data.attrs['Pixel_size'].item()
            except KeyError:
                pass
            else:
                crop_center = CropCenter(center_x_px, center_y_px)
                detector_pixel_geometry = PixelGeometry(pixel_size, pixel_size)
                probe_energy_eV = 1000.0 * photon_energy_keV  # noqa: N806

            metadata = DiffractionMetadata(
                num_patterns_per_array=[num_patterns],
                pattern_dtype=data.dtype,
                detector_distance_m=detector_distance_m,
                detector_extent=ImageExtent(detector_width, detector_height),
                detector_pixel_geometry=detector_pixel_geometry,
                crop_center=crop_center,
                probe_energy_eV=probe_energy_eV,
                exposure_time_s=exposure_time_s,
                file_path=file_path,
            )
            contents_tree = self._tree_builder.build(h5_file)
            array = H5DiffractionPatternArray(
                label=file_path.stem,
                indexes=numpy.arange(num_patterns),
                file_path=file_path,
                data_path=self._data_path,
            )
            return SimpleDiffractionDataset(metadata, contents_tree, [array])


def register_plugins(registry: PluginRegistry) -> None:
    registry.diffraction_file_readers.register_plugin(
        LYNXDiffractionFileReader(),
        simple_name='APS_LYNX',
        display_name='APS 31-ID-E LYNX Files (*.h5 *.hdf5)',
    )



from __future__ import annotations
from collections.abc import Sequence
from typing import Any

from PyQt5.QtCore import Qt, QAbstractTableModel, QModelIndex, QObject, QStringListModel
from PyQt5.QtWidgets import QDialogButtonBox, QTableView

from ptychodus.api.observer import Observable, Observer
from ptychodus.api.settings import SettingsRegistry

from ..view.settings import SettingsView
from .data import FileDialogFactory


class SettingsTableModel(QAbstractTableModel):
    def __init__(self, parent: QObject | None = None) -> None:
        super().__init__(parent)
        self._names: Sequence[str] = list()
        self._values: Sequence[str] = list()

    def set_names_and_values(self, names: Sequence[str], values: Sequence[str]) -> None:
        self.beginResetModel()
        self._names = names
        self._values = values
        self.endResetModel()

    def headerData(  # noqa: N802
        self,
        section: int,
        orientation: Qt.Orientation,
        role: int = Qt.ItemDataRole.DisplayRole,
    ) -> Any:
        if role == Qt.ItemDataRole.DisplayRole and orientation == Qt.Orientation.Horizontal:
            if section == 0:
                return 'Name'
            elif section == 1:
                return 'Value'

    def data(self, index: QModelIndex, role: int = Qt.ItemDataRole.DisplayRole) -> Any:
        if index.isValid() and role == Qt.ItemDataRole.DisplayRole:
            if index.column() == 0:
                return self._names[index.row()]
            elif index.column() == 1:
                return str(self._values[index.row()])

    def rowCount(self, parent: QModelIndex = QModelIndex()) -> int:  # noqa: N802
        return len(self._names)

    def columnCount(self, parent: QModelIndex = QModelIndex()) -> int:  # noqa: N802
        return 2


class SettingsController(Observer):
    def __init__(
        self,
        settings_registry: SettingsRegistry,
        view: SettingsView,
        table_view: QTableView,
        file_dialog_factory: FileDialogFactory,
    ) -> None:
        super().__init__()
        self._settings_registry = settings_registry
        self._view = view
        self._table_view = table_view
        self._file_dialog_factory = file_dialog_factory

        self._list_model = QStringListModel()
        self._table_model = SettingsTableModel()

        settings_registry.add_observer(self)

        view.list_view.setModel(self._list_model)
        selection_model = view.list_view.selectionModel()

        if selection_model is None:
            raise ValueError('selection_model is None!')
        else:
            selection_model.currentChanged.connect(self._update_view)

        self._table_view.setModel(self._table_model)

        open_button = view.button_box.button(QDialogButtonBox.StandardButton.Open)
        open_button.clicked.connect(self._open_settings)
        save_button = view.button_box.button(QDialogButtonBox.StandardButton.Save)
        save_button.clicked.connect(self._save_settings)

        self._sync_model_to_view()

    def _open_settings(self) -> None:
        file_path, _ = self._file_dialog_factory.get_open_file_path(
            self._view,
            'Open Settings',
            name_filters=self._settings_registry.get_open_file_filters(),
            selected_name_filter=self._settings_registry.get_open_file_filter(),
        )

        if file_path:
            self._settings_registry.open_settings(file_path)

    def _save_settings(self) -> None:
        file_path, _ = self._file_dialog_factory.get_save_file_path(
            self._view,
            'Save Settings',
            name_filters=self._settings_registry.get_save_file_filters(),
            selected_name_filter=self._settings_registry.get_save_file_filter(),
        )

        if file_path:
            self._settings_registry.save_settings(file_path)

    def _update_view(self, current: QModelIndex, previous: QModelIndex) -> None:
        if not current.isValid():
            return

        group_name = self._list_model.data(current, Qt.ItemDataRole.DisplayRole)
        group = self._settings_registry[group_name]
        names: list[str] = list()
        values: list[str] = list()

        for parameter_name, parameter in group.parameters().items():
            names.append(parameter_name)
            values.append(parameter.get_value_as_string())

        self._table_model.set_names_and_values(names, values)

    def _sync_model_to_view(self) -> None:
        self._list_model.setStringList(sorted(iter(self._settings_registry)))

        current = self._view.list_view.currentIndex()
        self._update_view(current, QModelIndex())

    def _update(self, observable: Observable) -> None:
        if observable is self._settings_registry:
            self._sync_model_to_view()



from .api import ObjectAPI, ProbeAPI, ProductAPI, ProbePositionsAPI, PositionsStreamingContext
from .core import ProductCore
from .item import ProductRepositoryItem, ProductRepositoryObserver
from .object_repository import ObjectRepository
from .probe_repository import ProbeRepository
from .repository import ProductRepository
from .scan_repository import ProbePositionsRepository
from .settings import ProductSettings

__all__ = [
    'ObjectAPI',
    'ObjectRepository',
    'PositionsStreamingContext',
    'ProbeAPI',
    'ProbeRepository',
    'ProductAPI',
    'ProductCore',
    'ProductRepository',
    'ProductRepositoryItem',
    'ProductRepositoryObserver',
    'ProductSettings',
    'ProbePositionsAPI',
    'ProbePositionsRepository',
]



from ptychodus.api.observer import Observable, Observer

from ..diffraction import AssembledDiffractionDataset
from .geometry import ProductGeometry
from .object import ObjectRepositoryItem
from .probe import ProbeRepositoryItem
from .probe_positions import ProbePositionsRepositoryItem


class ProductValidator(Observable, Observer):  # TODO display
    def __init__(
        self,
        dataset: AssembledDiffractionDataset,
        scan: ProbePositionsRepositoryItem,
        geometry: ProductGeometry,
        probe: ProbeRepositoryItem,
        object_: ObjectRepositoryItem,
    ) -> None:
        super().__init__()
        self._dataset = dataset
        self._scan = scan
        self._geometry = geometry
        self._probe = probe
        self._object = object_
        self._are_positions_valid = False
        self._are_probes_valid = False
        self._is_object_valid = False

    def are_positions_valid(self) -> bool:
        return self._are_positions_valid

    def are_probes_valid(self) -> bool:
        return self._are_probes_valid

    def is_object_valid(self) -> bool:
        return self._is_object_valid

    def _validate_scan(self) -> None:
        scan = self._scan.get_probe_positions()
        scan_indexes = set(point.index for point in scan)
        assembled_data = self._dataset.get_assembled_data()
        pattern_indexes = set(assembled_data.get_indexes())
        are_positions_valid_now = not scan_indexes.isdisjoint(pattern_indexes)

        if self._are_positions_valid != are_positions_valid_now:
            self._are_positions_valid = are_positions_valid_now
            self.notify_observers()

    def _validate_probes_and_object(self) -> None:
        has_validity_changed = False

        probes = self._probe.get_probes()
        are_probes_valid_now = self._geometry.is_probe_geometry_valid(probes.get_geometry())

        if self._are_probes_valid != are_probes_valid_now:
            self._are_probes_valid = are_probes_valid_now
            has_validity_changed = True

        object_ = self._object.get_object()
        is_object_valid_now = self._geometry.is_object_geometry_valid(object_.get_geometry())

        if self._is_object_valid != is_object_valid_now:
            self._is_object_valid = is_object_valid_now
            has_validity_changed = True

        if has_validity_changed:
            self.notify_observers()

    def _update(self, observable: Observable) -> None:
        if observable is self._dataset:
            self._validate_scan()
            self._validate_probes_and_object()
        elif observable is self._scan:
            self._validate_scan()
        elif observable is self._geometry:
            self._validate_probes_and_object()
        elif observable is self._probe:
            self._validate_probes_and_object()
        elif observable is self._object:
            self._validate_probes_and_object()



from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import (
    QFrame,
    QGridLayout,
    QGroupBox,
    QLabel,
    QMenu,
    QSizePolicy,
    QToolButton,
    QVBoxLayout,
    QWidget,
)


class GroupBoxWithPresets(QWidget):
    def __init__(self, title: str, parent: QWidget | None = None) -> None:
        super().__init__(parent)
        self._title_label = QLabel(title)

        self.presets_menu = QMenu()
        self._presets_button = QToolButton()
        self._presets_button.setText('Presets  ')
        self._presets_button.setToolButtonStyle(Qt.ToolButtonStyle.ToolButtonTextOnly)
        self._presets_button.setMenu(self.presets_menu)
        self._presets_button.setPopupMode(QToolButton.ToolButtonPopupMode.InstantPopup)

        self.contents = QWidget()
        frame_layout = QVBoxLayout()
        frame_layout.addWidget(self.contents)

        self._frame = QFrame()
        self._frame.setFrameShape(QFrame.Shape.StyledPanel)
        self._frame.setFrameShadow(QFrame.Shadow.Plain)
        self._frame.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Maximum)
        self._frame.setLayout(frame_layout)

        layout = QGridLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(self._title_label, 0, 0)
        layout.addWidget(self._presets_button, 0, 1, Qt.AlignmentFlag.AlignLeft)
        layout.addWidget(self._frame, 1, 0, 1, 2)
        layout.setColumnStretch(1, 1)
        self.setLayout(layout)


class BottomTitledGroupBox(QGroupBox):
    def __init__(self, title: str, parent: QWidget | None = None) -> None:
        super().__init__(title, parent)
        self.setStyleSheet("""
            QGroupBox::title {
                subcontrol-origin: padding;
                subcontrol-position: bottom center;
            }""")



from __future__ import annotations
from abc import abstractmethod
from dataclasses import dataclass
from pathlib import Path
import logging

import numpy

from ptychodus.api.common import RealArrayType
from ptychodus.api.parametric import ParameterGroup
from ptychodus.api.probe import (
    ProbeSequence,
    ProbeFileReader,
    ProbeGeometry,
    ProbeGeometryProvider,
    ComplexArrayType,
)

from .settings import ProbeSettings

logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class ProbeTransverseCoordinates:
    position_x_m: RealArrayType
    position_y_m: RealArrayType

    @property
    def position_r_m(self) -> RealArrayType:
        return numpy.hypot(self.position_x_m, self.position_y_m)


class ProbeSequenceBuilder(ParameterGroup):
    def __init__(self, settings: ProbeSettings, name: str) -> None:
        super().__init__()
        self._name = settings.builder.copy()
        self._name.set_value(name)
        self._add_parameter('name', self._name)

    def get_transverse_coordinates(self, geometry: ProbeGeometry) -> ProbeTransverseCoordinates:
        Y, X = numpy.mgrid[: geometry.height_px, : geometry.width_px]  # noqa: N806
        position_x_px = X - (geometry.width_px - 1) / 2
        position_y_px = Y - (geometry.height_px - 1) / 2

        position_x_m = position_x_px * geometry.pixel_width_m
        position_y_m = position_y_px * geometry.pixel_height_m

        return ProbeTransverseCoordinates(
            position_x_m=position_x_m,
            position_y_m=position_y_m,
        )

    def normalize(self, array: ComplexArrayType) -> ComplexArrayType:
        return array / numpy.sqrt(numpy.sum(numpy.square(numpy.abs(array))))

    def get_name(self) -> str:
        return self._name.get_value()

    def sync_to_settings(self) -> None:
        for parameter in self.parameters().values():
            parameter.sync_value_to_parent()

    @abstractmethod
    def copy(self) -> ProbeSequenceBuilder:
        pass

    @abstractmethod
    def build(self, geometry_provider: ProbeGeometryProvider) -> ProbeSequence:
        pass


class FromMemoryProbeBuilder(ProbeSequenceBuilder):
    def __init__(self, settings: ProbeSettings, probe: ProbeSequence) -> None:
        super().__init__(settings, 'from_memory')
        self._settings = settings
        self._probe = probe.copy()

    def copy(self) -> FromMemoryProbeBuilder:
        return FromMemoryProbeBuilder(self._settings, self._probe)

    def build(self, geometry_provider: ProbeGeometryProvider) -> ProbeSequence:
        probe_geometry = geometry_provider.get_probe_geometry()

        try:
            pixel_geometry = self._probe.get_pixel_geometry()
        except ValueError:
            pixel_geometry = probe_geometry.get_pixel_geometry()

        try:
            opr_weights = self._probe.get_opr_weights()
        except ValueError:
            opr_weights = None

        # TODO regrid probe as needed based on probe geometry from file/provider
        return ProbeSequence(
            self._probe.get_array(),
            opr_weights,
            pixel_geometry,
        )


class FromFileProbeBuilder(ProbeSequenceBuilder):
    def __init__(
        self, settings: ProbeSettings, file_path: Path, file_type: str, file_reader: ProbeFileReader
    ) -> None:
        super().__init__(settings, 'from_file')
        self._settings = settings
        self.file_path = settings.file_path.copy()
        self.file_path.set_value(file_path)
        self._add_parameter('file_path', self.file_path)
        self.file_type = settings.file_type.copy()
        self.file_type.set_value(file_type)
        self._add_parameter('file_type', self.file_type)
        self._file_reader = file_reader

    def copy(self) -> FromFileProbeBuilder:
        return FromFileProbeBuilder(
            self._settings,
            self.file_path.get_value(),
            self.file_type.get_value(),
            self._file_reader,
        )

    def build(self, geometry_provider: ProbeGeometryProvider) -> ProbeSequence:
        file_path = self.file_path.get_value()
        file_type = self.file_type.get_value()
        logger.debug(f'Reading "{file_path}" as "{file_type}"')

        try:
            probe_from_file = self._file_reader.read(file_path)
        except Exception as exc:
            raise RuntimeError(f'Failed to read "{file_path}"') from exc

        probe_geometry = geometry_provider.get_probe_geometry()

        try:
            pixel_geometry = probe_from_file.get_pixel_geometry()
        except ValueError:
            pixel_geometry = probe_geometry.get_pixel_geometry()

        try:
            opr_weights = probe_from_file.get_opr_weights()
        except ValueError:
            opr_weights = None

        # TODO regrid probe as needed based on probe geometry from file/provider
        return ProbeSequence(
            probe_from_file.get_array(),
            opr_weights,
            pixel_geometry,
        )
